import logging
import warnings

from bluesky import RunEngine as BlueskyRunEngine, suspenders
import databroker

from ._iconfig import load_config
from .preprocessors import inject_haven_md_wrapper
from .instrument.instrument_registry import registry

log = logging.getLogger(__name__)


# class RunEngine(BlueskyRunEngine):
#     def __init__(self, *args, connect_databroker=True, **kwargs):
#         super().__init__(*args, **kwargs)
#         if connect_databroker:
#             # Load the databroker catalog and set up data saving
#             catalog_name = load_config()["database"]["databroker"]["catalog"]
#             try:
#                 catalog = databroker.catalog[catalog_name]
#                 self.subscribe(catalog.v1.insert)
#             except Exception as e:
#                 msg = (
#                     f"Data are not being saved! Could not load databroker catalog: {e}"
#                 )
#                 log.error(msg)
#                 warnings.warn(msg)
#                 raise RuntimeError(msg)
#         # Add metadata pre-processor
#         self.preprocessors.append(inject_haven_md_wrapper)


catalog = None


def save_data(name, doc):
    # This is a hack around a problem with garbage collection
    # Has been fixed in main, maybe released in databroker v2?
    # Create the databroker callback if necessary
    global catalog
    if catalog is None:
        catalog = databroker.catalog["bluesky"]
    # Save the document
    catalog.v1.insert(name, doc)


def run_engine(connect_databroker=True) -> BlueskyRunEngine:
    RE = BlueskyRunEngine()
    # Install suspenders
    aps = registry.find("APS")
    # Suspend when shutter permit is disabled
    RE.install_suspender(
        suspenders.SuspendWhenChanged(
            signal=aps.shutter_permit,
            expected_value="PERMIT",
            allow_resume=True,
            sleep=3,
            tripped_message="Shutter permit revoked.",
        )
    )
    # Install databroker connection
    if connect_databroker:
        RE.subscribe(save_data)
    # Add preprocessors
    RE.preprocessors.append(inject_haven_md_wrapper)
    return RE



from qtpy import QtWidgets

from firefly import display


class XafsScanRegion:
    def __init__(self):
        self.setup_ui()

    def setup_ui(self):
        self.layout = QtWidgets.QHBoxLayout()
        # First energy box
        self.start_line_edit = QtWidgets.QLineEdit()
        self.start_line_edit.setPlaceholderText("Start…")
        self.layout.addWidget(self.start_line_edit)
        # Last energy box
        self.stop_line_edit = QtWidgets.QLineEdit()
        self.stop_line_edit.setPlaceholderText("Stop…")
        self.layout.addWidget(self.stop_line_edit)
        # Energy step box
        self.step_line_edit = QtWidgets.QLineEdit()
        self.step_line_edit.setPlaceholderText("Step…")
        self.layout.addWidget(self.step_line_edit)
        # K-space checkbox
        self.k_space_checkbox = QtWidgets.QCheckBox()
        self.k_space_checkbox.setText("K-space")
        self.k_space_checkbox.setEnabled(False)
        self.layout.addWidget(self.k_space_checkbox)
        # K-weight factor box, hidden at first
        self.k_weight_line_edit = QtWidgets.QLineEdit()
        self.k_weight_line_edit.setPlaceholderText("K-weight")
        self.k_weight_line_edit.setEnabled(False)
        self.layout.addWidget(self.k_weight_line_edit)
        # Connect the k-space enabled checkbox to the relevant signals
        self.k_space_checkbox.stateChanged.connect(self.k_weight_line_edit.setEnabled)

    def update_edge_enabled(self, is_checked: int):
        # Go back to real space if k-space was enabled
        if not is_checked:
            self.k_space_checkbox.setChecked(False)
        # Disabled the k-space checkbox
        self.k_space_checkbox.setEnabled(is_checked)


class XafsScanDisplay(display.FireflyDisplay):
    def customize_ui(self):
        self.reset_default_regions()
        # Connect the E0 checkbox to the E0 combobox
        self.ui.use_edge_checkbox.stateChanged.connect(self.edge_combo_box.setEnabled)

    def reset_default_regions(self):
        self.regions = []
        num_regions = 3
        self.ui.regions_spin_box.setValue(num_regions)
        self.add_regions(num_regions)

    def add_regions(self, num=1):
        for i in range(num):
            region = XafsScanRegion()
            self.ui.regions_layout.addLayout(region.layout)
            # Connect the E0 checkbox to each of the regions
            self.ui.use_edge_checkbox.stateChanged.connect(region.update_edge_enabled)
            # Save it to the list
            self.regions.append(region)

    def ui_filename(self):
        return "xafs_scan.ui"



from bluesky import suspenders
from ophyd import FormattedComponent as FCpt, EpicsSignal
from apstools.devices.shutters import ApsPssShutterWithStatus as Shutter

from .._iconfig import load_config
from .instrument_registry import registry


def load_shutters(config=None):
    if config is None:
        config = load_config()
    prefix = config["shutter"]["prefix"]
    for name, d in config["shutter"].items():
        if name == "prefix":
            continue
        # Calculate suitable PV values
        hutch = d["hutch"]
        acronym = "FES" if hutch == "A" else f"S{hutch}S"
        shutter = Shutter(
            prefix=f"{prefix}:{acronym}",
            open_pv=f"{prefix}:{acronym}_OPEN_EPICS.VAL",
            close_pv=f"{prefix}:{acronym}_CLOSE_EPICS.VAL",
            state_pv=f"{prefix}:{hutch}_BEAM_PRESENT",
            name=name,
            labels={"shutters"},
        )
        registry.register(shutter)



from ophyd import (
    Device,
    Component as Cpt,
    FormattedComponent as FCpt,
    EpicsSignal,
    EpicsSignalRO,
    EpicsMotor,
)

from .._iconfig import load_config
from .instrument_registry import registry


@registry.register
class NHQ203MChannel(Device):
    """A single channel on a controllable power supply."""

    ch_num: int

    # Device components
    potential = FCpt(
        EpicsSignal,
        name="potential",
        suffix="{prefix}:Volt{ch_num}_rbv",
        write_pv="{prefix}:SetVolt{ch_num}",
        tolerance=2,
    )
    current = FCpt(EpicsSignalRO, name="current", suffix="{prefix}:Curr{ch_num}_rbv")
    ramp_rate = FCpt(
        EpicsSignal,
        name="ramp_rate",
        suffix="{prefix}:RampSpeed{ch_num}",
        write_pv="{prefix}:RampSpeed{ch_num}_rbv",
    )
    status = FCpt(EpicsSignalRO, name="status", suffix="{prefix}:ModStatus{ch_num}_rbv")

    def __init__(self, prefix: str, ch_num: int, name: str, *args, **kwargs):
        self.ch_num = ch_num
        super().__init__(prefix=prefix, name=name, *args, **kwargs)


def load_power_supplies(config=None):
    if config is None:
        config = load_config()
    # Determine if any power supplies are available
    ps_configs = config.get("power_supply", {})
    for name, ps_config in ps_configs.items():
        # Do it once for each channel
        for ch_num in range(1, ps_config["n_channels"] + 1):
            this_name = f"{name}_ch{ch_num}"
            NHQ203MChannel(
                name=this_name,
                prefix=ps_config["prefix"],
                ch_num=ch_num,
                labels={"power_supplies"},
            )



from collections import namedtuple
from typing import Mapping

import matplotlib.pyplot as plt
import numpy as np
from lmfit.models import StepModel, GaussianModel
from bluesky import plans as bp

from ..instrument.instrument_registry import registry


def knife_scan(
    knife_motor,
    start: float,
    end: float,
    num: int,
    I0="I0",
    It="It",
    relative: bool = False,
    md: Mapping = {},
):
    """Plan to scan over a knife placed in the beam to measure its height.

    Parameters
    ==========
    knife_motor
      An ophyd device or name of a registered device that has a well
      defined edge attached to it.
    start
      Motor position at which to start scanning.
    end
      Motor position at which to finish scanning.
    num
      How many points to measure between *start* and *end*.
    I0
      Device or name of a registered ion chamber upstream from the
      knife.
    It
      Device or name of a registered ion chamber downstream from the
      knife.
    relative
      If true, *start* and *end* will be interpreted as relative to
      the current motor position.
    md
      Extra metadata to pass into the run engine.

    """
    md_ = dict(plan_name="knife_scan", **md)
    I0 = registry.find(name=I0)
    It = registry.find(name=It)
    knife_motor = registry.find(name=knife_motor)
    # Check for relative or absolute scan
    if relative:
        plan_func = bp.rel_scan
    else:
        plan_func = bp.scan
    # Compute the plan
    yield from plan_func([I0, It], knife_motor, start, end, num=num, md=md_)


def fit_step(x, y, plot=False, plot_derivative=False):
    """Extract beam properties from a step scan over the beam profile.

    Parameters
    ==========
    x
      Array of motor positions.
    y
      Array of measured values at each motor position
      (e.g. transmitted intensity).
    plot
      If true, plot the fitting results.
    plot_derivative
      If true, also fit and plot the derivative of the step scan.

    Returns
    =======
    properties
      A named tuple with the *size* and *position* of the beam, in
      units of *x*.

    """
    x = np.asarray(x)
    y = np.asarray(y)
    # Normalize data to start at zero
    y -= y[0]
    # Set up initial fitting parameters
    model = StepModel(form="erf")
    params = model.guess(y, x)
    half_max = (np.min(y) + np.max(y)) / 2
    center_idx = np.argmin(np.abs(y - half_max))
    center_guess = x[center_idx]
    params["center"].set(center_guess)
    # Invert the amplitude if we started with full transmission
    is_inverted = np.mean(y) < 0
    if is_inverted:
        params["amplitude"].set(-params["amplitude"].value)
    # Fit the model to the data
    result = model.fit(y, params=params, x=x)
    beam_position = result.values["center"]
    beam_hwhm = result.values["sigma"]
    # Do a fit of the derivative just for testing
    # Plot results
    if plot:
        # Plot knife scan and fit
        fig = plt.figure()
        result.plot(
            fig=fig,
            show_init=False,
            datafmt="x",
            xlabel="Knife position /µm",
            ylabel="Relative transmission",
        )
        ax_res, ax = fig.axes
        line_kw = dict(color="C0", ls=":")
        ax.axvline(beam_position, label="Beam position", **line_kw)
        ax.axhline(half_max, **line_kw)
        ax.text(beam_position, half_max, s=f"{beam_position:.2f}", color="C0")
        ax.axvline(beam_position - beam_hwhm, label="FWHM (y)", **line_kw)
        ax.axvline(beam_position + beam_hwhm, **line_kw)
        ax.legend()
    if plot and plot_derivative:
        # Fit derivative
        model = GaussianModel()
        dy = np.gradient(y, x)
        params = model.guess(dy, x)
        dresult = model.fit(dy, params=params, x=x)
        # Plot derivative of knife scan
        color2 = "C3"
        ax2 = ax.twinx()
        ax2.plot(x, dy, color=color2, label="Derivative", marker="x", ls="None")
        ax2.plot(x, dresult.eval(x=x), color=color2, alpha=0.5, label="Best fit")
        ax2.plot(
            x,
            np.gradient(result.eval(x=x), x),
            color=color2,
            alpha=0.5,
            ls="--",
            label="Step fit gradient",
        )
        dy_half_max = (np.min(dy) + np.max(dy)) / 2
        line_kw["color"] = color2
        ax2.axhline(dy_half_max, **line_kw)
        ax2.axvline(dresult.values["center"], **line_kw)
        dfwhm = dresult.values["fwhm"]
        ax2.axvline(
            dresult.values["center"] - dfwhm / 2, label="FWHM (dy/dx)", **line_kw
        )
        ax2.axvline(dresult.values["center"] + dfwhm / 2, **line_kw)
        ax2.legend()
        ax2.set_ylabel("Derivative /µm⁻")
    # Prepare results object
    Properties = namedtuple("Properties", ["position", "fwhm"])
    properties = Properties(position=beam_position, fwhm=2 * beam_hwhm)
    return properties



import pytest
from ophyd.sim import motor, motor1, motor2

from haven import set_energy, exceptions


def test_plan_messages():
    """Check that the right messages are getting produced."""
    plan = set_energy(energy=8400, positioners=[motor])
    msgs = list(plan)
    assert len(msgs) == 2
    msg0 = msgs[0]
    assert msg0.args == (8400,)
    assert msg0.obj.name == "motor"


def test_id_harmonic():
    """See if messages get emitted to change the ID harmonic at
    certain intervals.

    """
    plan = set_energy(
        energy=8400, harmonic=3, positioners=[motor1], harmonic_positioners=[motor2]
    )
    msgs = list(plan)
    # Check that a message exists to the ID harmonic
    assert len(msgs) == 4
    msg0 = msgs[0]
    assert msg0.args == (3,)
    assert msg0.obj.name == "motor2"


def test_id_harmonic_auto():
    plan = set_energy(
        energy=8400,
        harmonic="auto",
        positioners=[motor1],
        harmonic_positioners=[motor2],
    )
    msgs = list(plan)
    # Check that a message exists to the ID harmonic
    assert len(msgs) == 4
    msg0 = msgs[0]
    assert msg0.args == (1,)
    assert msg0.obj.name == "motor2"
    # Try again but with a 3rd harmonic
    plan = set_energy(
        energy=18400,
        harmonic="auto",
        positioners=[motor1],
        harmonic_positioners=[motor2],
    )
    msgs = list(plan)
    # Check that a message exists to the ID harmonic
    assert len(msgs) == 4
    msg0 = msgs[0]
    assert msg0.args == (3,)
    assert msg0.obj.name == "motor2"


def test_invalid_harmonic():
    plan = set_energy(
        energy=8400,
        harmonic="jabberwocky",
        positioners=[motor1],
        harmonic_positioners=[motor2],
    )
    with pytest.raises(exceptions.InvalidHarmonic):
        list(plan)



"""Holds ion chamber detector descriptions and assignments to EPICS PVs."""

from typing import Sequence
import logging

import epics
from ophyd import (
    Device,
    status,
    EpicsSignal,
    EpicsSignalRO,
    PVPositionerPC,
    PVPositioner,
    PseudoPositioner,
    PseudoSingle,
    Component as Cpt,
    FormattedComponent as FCpt,
    Kind,
)
from ophyd.ophydobj import OphydObject
from ophyd.pseudopos import pseudo_position_argument, real_position_argument

from .scaler_triggered import ScalerTriggered, ScalerSignal, ScalerSignalRO
from .instrument_registry import registry
from .._iconfig import load_config
from .. import exceptions


log = logging.getLogger(__name__)


__all__ = ["IonChamber", "load_ion_chambers"]


class SensitivityPositioner(PVPositionerPC):
    setpoint = Cpt(EpicsSignal, ".VAL", kind=(Kind.config | Kind.normal))
    readback = Cpt(EpicsSignal, ".VAL", kind=(Kind.config | Kind.normal))
    readback_string = Cpt(
        EpicsSignal, ".VAL", kind=(Kind.config | Kind.normal), string=True
    )


class SetAllPositioner(PVPositionerPC):
    setpoint = Cpt(EpicsSignal, "")
    readback = Cpt(EpicsSignal, "")


class SensitivityLevelPositioner(PseudoPositioner):
    values = [1, 2, 5, 10, 20, 50, 100, 200, 500]
    units = ["pA/V", "nA/V", "µA/V", "mA/V"]
    offset_difference = -3  # How many levels higher should the offset be

    sens_level = Cpt(
        PseudoSingle,
        limits=(0, 27),
        labels={"ion_chamber_sensitivities", "ion_chamber_gains"},
    )

    # Sensitivity settings
    sens_unit = Cpt(
        SensitivityPositioner, ":sens_unit", kind=Kind.config, settle_time=0.1
    )
    sens_value = Cpt(
        SensitivityPositioner, ":sens_num", kind=Kind.config, settle_time=0.1
    )
    offset_unit = Cpt(
        SensitivityPositioner, ":offset_unit", kind=Kind.config, settle_time=0.1
    )
    offset_value = Cpt(
        SensitivityPositioner, ":offset_num", kind=Kind.config, settle_time=0.1
    )
    set_all = Cpt(
        SetAllPositioner,
        ":init.PROC",
        kind=Kind.omitted,
        settle_time=0.1,
        limits=(0, 1),
    )

    def _level_to_num(self, level):
        return level % len(self.values)

    def _level_to_unit(self, level):
        return int(level / len(self.values))

    @pseudo_position_argument
    def forward(self, target_gain_level):
        "Given a target energy, transform to the desired target gain level."
        new_level = target_gain_level.sens_level
        new_offset = max(new_level + self.offset_difference, 0)
        real_position = self.RealPosition(
            sens_value=self._level_to_num(new_level),
            sens_unit=self._level_to_unit(new_level),
            offset_value=self._level_to_num(new_offset),
            offset_unit=self._level_to_unit(new_offset),
            set_all=1,
        )
        return real_position

    @real_position_argument
    def inverse(self, sensitivity):
        "Given a position in mono and ID energy, transform to the target energy."
        new_gain = sensitivity.sens_value + sensitivity.sens_unit * len(self.values)
        return self.PseudoPosition(sens_level=new_gain)


# @registry.register
class IonChamber(ScalerTriggered, Device):
    """An ion chamber at a spectroscopy beamline.

    Also includes the pre-amplifier as ``.pre_amp``.

    Parameters
    ==========
    prefix
      The PV prefix of the overall scaler.
    ch_num
      The number (1-index) of the channel on the scaler. 1 is the
      timer, so your channel number should start at 2.
    name
      The bluesky-compatible name for this device.
    preamp_prefix
      The process variable prefix to the pre-amp that controls this
      ion chamber (e.g. "25idc:SR01").
    scaler_prefix
      The process variable prefix for the scaler that measures this
      ion chamber.

    Attributes
    ==========
    ch_num
      The channel number on the scaler, starting at 2 (1 is the timer).
    count
      The trigger to count scaler pulses.
    raw_counts
      The counts coming from the scaler without any correction.
    volts
      The volts produced by the pre-amp, calculated from scaler
      counts.
    exposure_time
      Positioner for setting the count time on the scaler.
    sensitivity
      The positioner for changing the pre-amp gain/sensitivity.

    """

    ch_num: int = 0
    ch_char: str
    count: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}.CNT", trigger_value=1, kind=Kind.omitted
    )
    description: OphydObject = FCpt(
        EpicsSignalRO, "{prefix}.NM{ch_num}", kind=Kind.config
    )
    raw_counts: OphydObject = FCpt(
        ScalerSignalRO, "{prefix}.S{ch_num}", kind=Kind.normal
    )
    offset: OphydObject = FCpt(
        ScalerSignalRO, "{prefix}_{offset_suffix}", kind=Kind.config
    )
    net_counts: OphydObject = FCpt(
        ScalerSignalRO, "{prefix}_netA.{ch_char}", kind=Kind.hinted
    )
    volts: OphydObject = FCpt(
        ScalerSignalRO, "{prefix}_calc{ch_num}.VAL", kind=Kind.normal
    )
    exposure_time: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}.TP", kind=Kind.normal
    )
    sensitivity: OphydObject = FCpt(
        SensitivityLevelPositioner, "{preamp_prefix}", kind=Kind.config
    )
    auto_count: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}.CONT", kind=Kind.omitted
    )
    record_dark_current: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}_offset_start.PROC", kind=Kind.omitted
    )
    record_dark_time: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}_offset_time.VAL", kind=Kind.config
    )

    _default_read_attrs = [
        "raw_counts",
        "volts",
        "exposure_time",
        "net_counts",
    ]

    def __init__(
        self,
        prefix: str,
        ch_num: int,
        name: str,
        preamp_prefix: str = None,
        scaler_prefix: str = None,
        *args,
        **kwargs,
    ):
        # Set up the channel number for this scaler channel
        if ch_num < 1:
            raise ValueError(f"Scaler channels must be greater than 0: {ch_num}")
        self.ch_num = ch_num
        self.ch_char = self.num_to_char(ch_num)
        # Determine which prefix to use for the scaler
        if scaler_prefix is not None:
            self.scaler_prefix = scaler_prefix
        else:
            self.scaler_prefix = prefix
        # Save an epics path to the preamp
        if preamp_prefix is None:
            preamp_prefix = prefix
        self.preamp_prefix = preamp_prefix
        # Determine the offset PV since it follows weird numbering conventions
        calc_num = int((self.ch_num - 1) / 4)
        calc_char = self.num_to_char(((self.ch_num - 1) % 4) + 1)
        self.offset_suffix = f"offset{calc_num}.{calc_char}"
        # Initialize all the other Device stuff
        super().__init__(prefix=prefix, name=name, *args, **kwargs)
        # Set signal values to stage
        self.stage_sigs[self.auto_count] = 0

    def num_to_char(self, num):
        char = chr(64 + num)
        return char

    def change_sensitivity(self, step: int) -> status.StatusBase:
        """Change the gain on the pre-amp by the given number of steps.

        Parameters
        ==========
        step
          How many levels to change the sensitivity. Positive numbers
          increase the gain, negative numbers decrease the gain.

        Returns
        =======
        status.StatusBase
          The status that will be marked complete once the sensitivity
          is changed.

        """
        new_sens_level = self.sensitivity.sens_level.readback.get() + step
        try:
            status = self.sensitivity.sens_level.set(new_sens_level)
        except ValueError as e:
            raise exceptions.GainOverflow(f"{self.name} -> {e}")
        return status

    def increase_gain(self) -> Sequence[status.StatusBase]:
        """Increase the gain (descrease the sensitivity) of the ion chamber's
        pre-amp.

        Returns
        =======
        Sequence[status.StatusBase]
          Ophyd status objects for the value and gain of the
          sensitivity in the pre-amp.

        """
        return self.change_sensitivity(-1)

    def decrease_gain(self) -> Sequence[status.StatusBase]:
        """Decrease the gain (increase the sensitivity) of the ion chamber's
        pre-amp.

        Returns
        =======
        Sequence[status.StatusBase]
          Ophyd status objects for the value and gain of the
          sensitivity in the pre-amp.

        """
        return self.change_sensitivity(1)


def load_ion_chambers(config=None):
    # Load IOC prefixes from the config file
    if config is None:
        config = load_config()
    vme_ioc = config["ion_chamber"]["scaler"]["ioc"]
    scaler_record = config["ion_chamber"]["scaler"]["record"]
    scaler_pv_prefix = f"{vme_ioc}:{scaler_record}"
    preamp_ioc = config["ion_chamber"]["preamp"]["ioc"]
    ion_chambers = []
    # Loop through the configuration sections and create the ion chambers
    for ch_num in config["ion_chamber"]["scaler"]["channels"]:
        # Determine ion_chamber configuration
        preamp_prefix = f"{preamp_ioc}:SR{ch_num-1:02}"
        desc_pv = f"{scaler_pv_prefix}.NM{ch_num}"
        # Only use this ion chamber if it has a name
        name = epics.caget(desc_pv, as_string=True)
        if name == "":
            log.info(f"Skipping unnamed ion chamber: {desc_pv}")
            continue
        if name is None:
            name = "???"
        # Create the ion chamber
        ic = IonChamber(
            prefix=scaler_pv_prefix,
            ch_num=ch_num,
            name=name,
            preamp_prefix=preamp_prefix,
            labels={"ion_chambers"},
        )
        registry.register(ic)
        log.info(f"Created ion chamber: {ic.name} ({ic.prefix}, ch {ic.ch_num})")
        ion_chambers.append(ic)
    return ion_chambers



from bluesky import plans as bp, plan_stubs as bps

from .shutters import open_shutters, close_shutters
from ..instrument.instrument_registry import registry


def record_dark_current(ion_chambers, shutters, time):
    """Record the dark current on the ion chambers.

    - Close shutters
    - Record ion chamber dark current
    - Open shutters

    Parameters
    ==========
    ion_chambers
      Ion chamber devices or names.
    shutters
      Shutter devices or names.

    """
    yield from close_shutters(shutters)
    # Measure the dark current
    ion_chambers = registry.findall(ion_chambers)
    # This is a big hack, we need to come back and just accept the current integration time
    old_times = [ic.exposure_time.get() for ic in ion_chambers]
    time_args = [obj for ic in ion_chambers for obj in (ic.record_dark_time, time)]
    yield from bps.mv(*time_args)
    mv_args = [obj for ic in ion_chambers for obj in (ic.record_dark_current, 1)]
    triggers = [ic.record_dark_current for ic in ion_chambers]
    yield from bps.mv(*mv_args)
    yield from bps.sleep(time)
    time_args = [
        obj for (ic, t) in zip(ion_chambers, old_times) for obj in (ic.exposure_time, t)
    ]
    yield from bps.mv(*time_args)
    # Open shutters again
    yield from open_shutters(shutters)



import logging
import os
import time
import time_machine
import pytz
from zoneinfo import ZoneInfo
import datetime as dt
from datetime import datetime

import epics
import pytest
from ophyd.sim import motor1, SynAxis, make_fake_device
from ophyd import EpicsMotor, Signal, Component as Cpt
from haven import (
    save_motor_position,
    registry,
    get_motor_position,
    list_motor_positions,
    recall_motor_position,
    list_current_motor_positions,
    HavenMotor,
)

log = logging.getLogger(__name__)

# Use a timezone we're not likely to be in for testing tz-aware behavior
fake_time = dt.datetime(2022, 8, 19, 19, 10, 51, tzinfo=ZoneInfo("Asia/Taipei"))

IOC_timeout = 40  # Wait up to this many seconds for the IOC to be ready

motor_prefix = "255idVME:"


class FakeHavenMotor(SynAxis):
    user_offset = Cpt(Signal, value=0, kind="config")


@pytest.fixture
def sim_motor_registry(sim_registry):
    # Create the motors
    motor1 = FakeHavenMotor(name="SLT V Upper")
    sim_registry.register(motor1)
    motor2 = FakeHavenMotor(name="SLT V Lower")
    sim_registry.register(motor2)
    yield sim_registry


@time_machine.travel(fake_time, tick=True)
def test_save_motor_position_by_device(mongodb):
    # Check that no entry exists before saving it
    result = mongodb.motor_positions.find_one({"name": motor1.name})
    assert result is None
    # Create motor devices
    motorA = SynAxis(name="Motor A")
    motorB = SynAxis(name="Motor B")
    motorA.wait_for_connection()
    motorB.wait_for_connection()
    # Move to some other motor position so we can tell it saved the right one
    motorA.set(11.0).wait(timeout=10)
    motorB.set(23.0).wait(timeout=10)
    # Save the current motor position
    save_motor_position(
        motorA, motorB, name="Sample center", collection=mongodb.motor_positions
    )
    # Check that the motors got saved
    result = mongodb.motor_positions.find_one({"name": "Sample center"})
    assert result is not None
    assert len(result["motors"]) == 2
    result_A = [r for r in result["motors"] if r["name"] == motorA.name][0]
    result_B = [r for r in result["motors"] if r["name"] == motorB.name][0]
    assert result_A["name"] == motorA.name
    assert result_A["readback"] == 11.0
    assert result_B["readback"] == 23.0
    # Check that the timestamp was saved (accurate to within a second)
    assert result["savetime"] == pytest.approx(time.time(), abs=1)


@time_machine.travel(fake_time, tick=True)
def test_save_motor_position_by_name(mongodb, sim_registry):
    # Check that no entry exists before saving it
    result = mongodb.motor_positions.find_one({"name": motor1.name})
    assert result is None
    # Get our simulated motors into the device registry
    motorA = FakeHavenMotor(name="Motor A")
    motorB = FakeHavenMotor(name="Motor B")
    motorA.wait_for_connection(timeout=20)
    motorB.wait_for_connection(timeout=20)
    # Register the new motors with the Haven instrument registry
    sim_registry.register(motorA)
    sim_registry.register(motorB)
    # Move to some other motor position so we can tell it saved the right one
    motorA.set(11.0).wait()
    motorA.user_offset.set(1.5).wait()
    motorB.set(23.0).wait()
    time.sleep(0.1)
    # Save the current motor position
    save_motor_position(
        "Motor A", "Motor B", name="Sample center", collection=mongodb.motor_positions
    )
    # Check that the motors got saved
    result = mongodb.motor_positions.find_one({"name": "Sample center"})
    assert result is not None
    assert len(result["motors"]) == 2
    result_A = [r for r in result["motors"] if r["name"] == motorA.name][0]
    result_B = [r for r in result["motors"] if r["name"] == motorB.name][0]
    assert result_A["name"] == motorA.name
    assert result_A["readback"] == 11.0
    assert result_B["readback"] == 23.0
    assert result_A["offset"] == 1.5
    # Check that the metadata saved
    assert result["savetime"] == pytest.approx(time.time(), abs=1)


def test_get_motor_position_by_uid(mongodb):
    uid = str(mongodb.motor_positions.find_one({"name": "Good position A"})["_id"])
    result = get_motor_position(uid=uid, collection=mongodb.motor_positions)
    assert result.name == "Good position A"
    assert result.motors[0].name == "SLT V Upper"
    assert result.motors[0].readback == 510.5


def test_get_motor_position_by_name(mongodb):
    result = get_motor_position(
        name="Good position A", collection=mongodb.motor_positions
    )
    assert result.name == "Good position A"
    assert result.motors[0].name == "SLT V Upper"
    assert result.motors[0].readback == 510.5


def test_get_motor_position_exceptions(mongodb):
    # Fails when no query params are given
    with pytest.raises(TypeError):
        get_motor_position(collection=mongodb.motor_positions)


def test_recall_motor_position(mongodb, sim_motor_registry):
    # Re-set the previous value
    uid = str(mongodb.motor_positions.find_one({"name": "Good position A"})["_id"])
    plan = recall_motor_position(uid=uid, collection=mongodb.motor_positions)
    messages = list(plan)
    # Check the plan output
    msg0 = messages[0]
    assert msg0.obj.name == "SLT V Upper"
    assert msg0.args[0] == 510.5
    msg1 = messages[1]
    assert msg1.obj.name == "SLT V Lower"
    assert msg1.args[0] == -211.93


@time_machine.travel(fake_time, tick=True)
def test_list_motor_positions(mongodb, capsys):
    # Do the listing
    list_motor_positions(collection=mongodb.motor_positions)
    # Check stdout for printed motor positions
    captured = capsys.readouterr()
    assert len(captured.out) > 0
    uid = str(mongodb.motor_positions.find_one({"name": "Good position A"})["_id"])
    timestamp = "2022-08-19 19:10:51"
    expected = (
        f'\n\033[1mGood position A\033[0m (uid="{uid}", timestamp={timestamp})\n'
        "┣━SLT V Upper: 510.5, offset: 0.0\n"
        "┗━SLT V Lower: -211.93, offset: None\n"
    )
    assert captured.out == expected


def test_motor_position_e2e(mongodb, ioc_motor):
    """Check that a motor position can be saved, then recalled using
    simulated IOC.

    """
    # Create an epics motor for setting values manually
    pv = ioc_motor.pvs["m1"]
    motor1 = EpicsMotor(pv, name="SLT V Upper")
    motor1.wait_for_connection(timeout=20)
    assert motor1.connected
    registry.register(motor1)
    registry.find(name="SLT V Upper")
    epics.caput(pv, 504.6)
    assert epics.caget(pv, use_monitor=False) == 504.6
    time.sleep(0.1)
    assert motor1.get(use_monitor=False).user_readback == 504.6
    # Save motor position
    uid = save_motor_position(
        motor1,
        name="starting point",
        collection=mongodb.motor_positions,
    )
    # Change to a different value
    epics.caput(pv, 520)
    time.sleep(0.1)
    assert epics.caget(pv, use_monitor=False) == 520
    assert motor1.get(use_monitor=False).user_readback == 520
    # Recall the saved position and see if it complies
    plan = recall_motor_position(uid=uid, collection=mongodb.motor_positions)
    msg = next(plan)
    assert msg.obj.name == "SLT V Upper"
    assert msg.args[0] == 504.6


@time_machine.travel(fake_time, tick=True)
def test_list_current_motor_positions(mongodb, capsys):
    # Get our simulated motors into the device registry
    with capsys.disabled():
        motorA = FakeHavenMotor(name="Motor A")
        motorB = FakeHavenMotor(name="Motor B")
        motorA.wait_for_connection()
        motorB.wait_for_connection()
        # Move to some other motor position so we can tell it saved the right one
        motorA.set(11.0).wait()
        motorA.user_offset.set(1.5).wait()
        motorB.set(23.0).wait()
    # List the current motor position
    list_current_motor_positions(motorA, motorB, name="Current motor positions")
    # Check stdout for printed motor positions
    captured = capsys.readouterr()
    assert len(captured.out) > 0
    timestamp = fake_time.strftime("%Y-%m-%d %H:%M:%S")
    timestamp = "2022-08-19 19:10:51"
    expected = (
        f"\n\033[1mCurrent motor positions\033[0m (timestamp={timestamp})\n"
        "┣━Motor A: 11.0, offset: 1.5\n"
        "┗━Motor B: 23.0, offset: 0.0\n"
    )
    assert captured.out == expected



from typing import Mapping

from ophyd import sim

from .instrument_registry import registry as default_registry, InstrumentRegistry
from .energy_positioner import load_energy_positioner
from .motor import load_all_motors
from .ion_chamber import load_ion_chambers
from .fluorescence_detector import load_fluorescence_detectors
from .monochromator import load_monochromator
from .camera import load_cameras
from .shutter import load_shutters
from .stage import load_stages
from .aps import load_aps
from .power_supply import load_power_supplies
from .xray_source import load_xray_sources
from .area_detector import load_area_detectors
from .slits import load_slits
from .lerix import load_lerix_spectrometers
from .heater import load_heaters
from .._iconfig import load_config


__all__ = ["load_instrument"]


def load_instrument(
    registry: InstrumentRegistry = default_registry, config: Mapping = None
):
    """Load the beamline instrumentation into an instrument registry.

    This function will reach out and query various IOCs for motor
    information based on the information in *config* (see
    ``iconfig_default.toml`` for examples). Based on the
    configuration, it will create Ophyd devices and register them with
    *registry*.

    Parameters
    ==========
    registry:
      The registry into which the ophyd devices will be placed.
    config:
      The beamline configuration read in from TOML files. Mostly
    useful for testing.

    """
    # Clear out any existing registry entries
    registry.clear()
    # Make sure we have the most up-to-date configuration
    load_config.cache_clear()
    # Load the configuration
    if config is None:
        config = load_config()
    # Import each device type for the instrument
    load_simulated_devices(config=config)
    load_energy_positioner(config=config)
    load_all_motors(config=config)
    load_monochromator(config=config)
    load_cameras(config=config)
    load_shutters(config=config)
    load_stages(config=config)
    load_power_supplies(config=config)
    load_slits(config=config)
    load_heaters(config=config)
    # Detectors
    load_ion_chambers(config=config)
    load_fluorescence_detectors(config=config)
    load_area_detectors(config=config)
    load_lerix_spectrometers(config=config)
    # Facility-related devices
    load_aps(config=config)
    load_xray_sources(config=config)


def load_simulated_devices(config={}):
    # Motors
    motor = sim.SynAxis(name="sim_motor", labels={"motors"})
    default_registry.register(motor)
    # Detectors
    detector = sim.SynGauss(
        name="sim_detector",
        motor=motor,
        motor_field="sim_motor",
        center=0,
        Imax=1,
    )
    default_registry.register(detector)



import logging
from typing import Sequence, Mapping, Optional

from pydm.widgets.channel import PyDMChannel
from haven.instrument.ion_chamber import IonChamber
from haven import load_config, exceptions, registry

from firefly import display

log = logging.getLogger(__name__)


class VoltmeterDisplay(display.FireflyDisplay):
    """PyDM display for showing a single voltmeter.

    Parameters
    ==========
    device
      An ophyd Device object for the ion chamber. If ``None``, then
      the device will be retrieved from the haven instrument registry
      based on its name in *macros["IC"]*.
    args
      Additional argument passed to the parent display class.
    macros
      Macros used to build the display (described below).

    Macros
    ======
    IC
      The name of the ophyd device describing the ion chamber
      associated with this voltmeter. The Device object itself will be
      retrieved from the haven registry if *device* is ``None``
      (default).

    """

    _device: IonChamber = None
    gain_values = [1, 2, 5, 10, 20, 50, 100, 200, 500]
    gain_units = ["pA/V", "nA/V", "µA/V", "mA/V"]
    gain = float("nan")
    gain_unit = "A"
    last_voltage = 0

    def __init__(
        self,
        device: IonChamber = None,
        args: Optional[Sequence] = None,
        macros: Mapping = {},
        **kwargs,
    ):
        self._device = device
        super().__init__(macros=macros, args=args, **kwargs)

    def customize_device(self):
        # Find and store the hardware device
        device_name = self.macros().get("IC")
        if self._device is None:
            try:
                self._device = registry.find(name=device_name)
            except exceptions.ComponentNotFound:
                log.warning(f"No device loaded for voltmeter: {self.macros()}.")
            else:
                log.debug(f"Voltmeter found device: {self._device}")
        # Setup a pydm channel to monitor the gain on the pre-amplifier
        self._ch_gain_value = PyDMChannel(
            self.ui.sens_num_label.channel, value_slot=self.update_gain
        )
        self._ch_gain_unit = PyDMChannel(
            self.ui.sens_unit_label.channel, value_slot=self.update_gain_unit
        )
        self._ch_voltage = PyDMChannel(
            self.ui.ion_chamber_label.channel, value_slot=self.update_current
        )
        for ch in [self._ch_voltage, self._ch_gain_value, self._ch_gain_unit]:
            ch.connect()

    def update_gain(self, new_gain_idx):
        self.gain = self.gain_values[new_gain_idx]
        log.debug(f"Updated gain value: {self.gain}")
        self.update_current(self.last_voltage)

    def update_gain_unit(self, new_unit_idx):
        unit = self.gain_units[new_unit_idx]
        self.gain_unit = unit.split("/")[0]
        log.debug(f"Updated gain unit: {self.gain_unit}")
        self.update_current(self.last_voltage)

    def update_current(self, voltage):
        self.last_voltage = voltage
        text = f"({voltage * self.gain:.02f} {self.gain_unit})"
        log.debug(f"New current text: {text}")
        self.ui.ion_chamber_current.setText(text)

    def customize_ui(self):
        # Gain adjustment buttons
        if self._device is not None:
            self.ui.gain_up_button.clicked.connect(self._device.increase_gain)
            self.ui.gain_down_button.clicked.connect(self._device.decrease_gain)
        else:
            log.warning("Cannot attach voltmeter buttons to device.")

    def ui_filename(self):
        return "voltmeter.ui"



from enum import IntEnum
from pathlib import Path
import datetime as dt
import subprocess
import logging
import os

import haven

# from pydm.data_plugins.epics_plugin import EPICSPlugin
from pydm.widgets.channel import PyDMChannel
from qtpy.QtGui import QColor
from qtpy.QtCore import Slot

from firefly import display, FireflyApplication


log = logging.getLogger(__name__)


class DetectorStates(IntEnum):
    IDLE = 0
    ACQUIRE = 1
    READOUT = 2
    CORRECT = 3
    SAVING = 4
    ABORTING = 5
    ERROR = 6
    WAITING = 7
    INITIALIZING = 8
    DISCONNECTED = 9
    ABORTED = 10


class AcquireStates(IntEnum):
    DONE = 0
    ACQUIRE = 1


class CameraDisplay(display.FireflyDisplay):
    prefix: str = ""
    properties_file: Path = Path("~/EPICS_AD_Viewer.properties").expanduser()
    _camera_state: int = DetectorStates.IDLE
    _camera_connected: bool = False

    def __init__(self, *, args=None, macros={}, **kwargs):
        self.prefix = macros.get("PREFIX", "")
        super().__init__(args=args, macros=macros, **kwargs)
        # Disconnect previous channels to the indicator
        byte = self.camera_status_indicator
        for ch in byte.channels():
            ch.disconnect()
        # Channel for watching the detector state
        self.detector_state = PyDMChannel(
            address=self.camera_status_label.channel,
            connection_slot=self.update_camera_connection,
            value_slot=self.update_camera_state,
        )
        self.detector_state.connect()
        byte._channels.append(self.detector_state)
        # Color for various states
        self._ioc_disconnected_color = QColor(255, 255, 255)
        self._detector_disconnected_color = QColor(255, 0, 0)
        self._idle_color = QColor(0, 255, 0)
        self._acquire_color = QColor(255, 255, 0)

    def customize_ui(self):
        # Connect button for opening the individual camera viewers
        app = FireflyApplication.instance()
        try:
            action = app.camera_actions[self.macros()["CAMERA"]]
        except KeyError:
            pass
        else:
            self.ui.viewer_button.clicked.connect(action.trigger)

    def ui_filename(self):
        return "camera.ui"

    @Slot(int)
    def update_camera_state(self, new_state):
        self._camera_state = new_state
        self.update_status_indicators()

    @Slot(bool)
    def update_camera_connection(self, new_state):
        self._camera_connected = new_state
        self.update_status_indicators()

    def update_status_indicators(self):
        # Retrieve widgets we're going to change
        bit = self.camera_status_indicator._indicators[0]
        lbl = self.camera_status_label
        # Determine the new state of the camera
        ioc_is_disconnected = not self._camera_connected
        new_state = self._camera_state
        camera_is_idle = new_state == DetectorStates.IDLE
        camera_is_acquiring = new_state == DetectorStates.ACQUIRE
        # Update the widgets
        if ioc_is_disconnected:
            # IOC is not running or not available
            bit.setColor(self._ioc_disconnected_color)
            lbl.setVisible(False)
        elif camera_is_idle:
            # Camera is idle
            bit.setColor(self._idle_color)
            lbl.setVisible(False)
        elif camera_is_acquiring:
            bit.setColor(self._acquire_color)
            lbl.setVisible(False)
        else:
            # Camera is disconnected or in an otherwise unexpected state
            bit.setColor(self._detector_disconnected_color)
            lbl.setVisible(True)



from .ion_chamber import IonChamber  # noqa: F401
from .motor import HavenMotor  # noqa: F401
from .monochromator import Monochromator  # noqa: F401
from .instrument_registry import InstrumentRegistry, registry  # noqa: F401



import time
from typing import Optional
import logging
import warnings

from qtpy.QtCore import QThread, QObject, Signal, Slot, QTimer
from bluesky_queueserver_api.zmq import REManagerAPI
from bluesky_queueserver_api import BPlan, comm_base

from haven import load_config


log = logging.getLogger()


class QueueClientThread(QThread):
    timer: QTimer

    def __init__(self, *args, client, **kwargs):
        self.client = client
        super().__init__(*args, **kwargs)
        # Timer for polling the queueserver
        self.timer = QTimer()
        self.timer.timeout.connect(self.client.update)
        self.timer.start(1000)

    def quit(self, *args, **kwargs):
        self.timer.stop()
        # del self.timer
        # del self.client
        super().quit(*args, **kwargs)


class QueueClient(QObject):
    api: REManagerAPI
    _last_queue_length: Optional[int] = None
    last_update: float = -1
    timeout: float = 1

    # Signals responding to queue changes
    state_changed = Signal()
    length_changed = Signal(int)

    def __init__(self, *args, api, **kwargs):
        self.api = api
        super().__init__(*args, **kwargs)

    def update(self):
        now = time.time()
        if now >= self.last_update + self.timeout:
            if not load_config()["beamline"]["is_connected"]:
                log.warning("Beamline not connected, skipping queue client update.")
                self.timeout = 60  # Just update every 1 minute
                self.last_update = now
                return
            log.debug("Updating queue client.")
            try:
                self.check_queue_length()
            except comm_base.RequestTimeoutError as e:
                # If we can't reach the server, wait for a minute and retry
                self.timeout = min(60, self.timeout * 2)
                log.warn(str(e))
                warnings.warn(str(e))
                log.info(f"Retrying in {self.timeout} seconds.")
            else:
                # Update succeeded, so wait for a second
                self.timeout = 1
            finally:
                self.last_update = now

    @Slot(bool)
    def request_pause(self, defer: bool = True):
        """Ask the queueserver run engine to pause.

        Parameters
        ==========
        defer
          If true, the run engine will be paused at the next
          checkpoint. Otherwise, it will pause now.

        """
        option = "deferred" if defer else "immediate"
        self.api.re_pause(option=option)

    @Slot(object)
    def add_queue_item(self, item):
        log.info(f"Client adding item to queue: {item}")
        result = self.api.item_add(item=item)
        if result["success"]:
            log.info(f"Item added. New queue length: {result['qsize']}")
            self.length_changed.emit(result["qsize"])
        else:
            log.error(f"Did not add queue item to queue: {result}")
            raise RuntimeError(result)

    @Slot()
    def start_queue(self):
        result = self.api.queue_start()
        log.debug(f"Started queue: {result}")

    @Slot()
    def check_queue_length(self):
        queue = self.api.queue_get()
        queue_length = len(queue["items"])
        log.debug(f"Queue length updated: {queue_length}")
        self.length_changed.emit(queue_length)
        self._last_queue_length = queue_length



import pytest

from ophyd import sim, Device, EpicsMotor

from haven import exceptions
from haven.instrument import InstrumentRegistry


def test_register_component():
    # Prepare registry
    reg = InstrumentRegistry()
    # Create an unregistered component
    cpt = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    # Make sure the component doesn't get found without being registered
    with pytest.raises(exceptions.ComponentNotFound):
        list(reg.findall(label="ion_chamber"))
    with pytest.raises(exceptions.ComponentNotFound):
        list(reg.findall(name="I0"))
    # Now register the component
    cpt = reg.register(cpt)
    # Confirm that it's findable by label
    results = reg.findall(label="ion_chamber")
    assert cpt in results
    # Config that it's findable by name
    results = reg.findall(name="I0")
    assert cpt in results


def test_find_missing_components():
    """Test that registry raises an exception if no matches are found."""
    reg = InstrumentRegistry()
    cpt = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    reg.register(cpt)
    # Now make sure a different query still returns no results
    with pytest.raises(exceptions.ComponentNotFound):
        reg.findall(label="spam")


def test_find_allow_missing_components():
    """Test that registry tolerates missing components with the
    *allow_none* argument.

    """
    reg = InstrumentRegistry()
    # Get some non-existent devices and check that the right nothing is returned
    assert list(reg.findall(label="spam", allow_none=True)) == []
    assert reg.find(name="eggs", allow_none=True) is None


def test_exceptions():
    reg = InstrumentRegistry()
    reg.register(Device("", name="It"))
    # Test if a non-existent labels throws an exception
    with pytest.raises(exceptions.ComponentNotFound):
        reg.find(label="spam")


def test_as_class_decorator():
    reg = InstrumentRegistry()
    # Create a dummy decorated class
    IonChamber = type("IonChamber", (Device,), {})
    IonChamber = reg.register(IonChamber)
    # Instantiate the class
    IonChamber("PV_PREFIX", name="I0", labels={"ion_chamber"})
    # Check that it gets retrieved
    result = reg.find(label="ion_chamber")
    assert result.prefix == "PV_PREFIX"
    assert result.name == "I0"


def test_find_component():
    # Prepare registry
    reg = InstrumentRegistry()
    # Create an unregistered component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    cptB = sim.SynGauss(
        "It",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    # Register the components
    reg.register(cptA)
    reg.register(cptB)
    # Only one match should work fine
    result = reg.find(name="I0")
    assert result is cptA
    # Multiple matches should raise an exception
    with pytest.raises(exceptions.MultipleComponentsFound):
        result = reg.find(label="ion_chamber")


def test_find_name_by_dot_notation():
    # Prepare registry
    reg = InstrumentRegistry()
    # Create a simulated component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    reg.register(cptA)
    # Only one match should work fine
    result = reg.find(name="I0.val")
    assert result is cptA.val


def test_find_labels_by_dot_notation():
    # Prepare registry
    reg = InstrumentRegistry()
    # Create a simulated component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    reg.register(cptA)
    # Only one match should work fine
    result = reg.find(label="ion_chamber.val")
    assert result is cptA.val


def test_find_any():
    # Prepare registry
    reg = InstrumentRegistry()
    # Create an unregistered component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    cptB = sim.SynGauss(
        "It",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    cptC = sim.SynGauss(
        "ion_chamber",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
    )
    cptD = sim.SynGauss(
        "sample motor",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    # Register the components
    reg.register(cptA)
    reg.register(cptB)
    reg.register(cptC)
    reg.register(cptD)
    # Only one match should work fine
    result = reg.findall(any_of="ion_chamber")
    assert cptA in result
    assert cptB in result
    assert cptC in result
    assert cptD not in result


def test_find_by_device():
    """The registry should just return the device itself if that's what is passed."""
    # Prepare registry
    reg = InstrumentRegistry()
    # Register a component
    cptD = sim.SynGauss(
        "sample motor",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    reg.register(cptD)
    # Pass the device itself to the find method
    result = reg.find(cptD)
    assert result is cptD


def test_find_by_list_of_names():
    """Will the findall() method handle lists of things to look up."""
    # Prepare registry
    reg = InstrumentRegistry()
    # Register a component
    cptA = sim.SynGauss(
        "sample motor A",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    cptB = sim.SynGauss(
        "sample motor B",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    cptC = sim.SynGauss(
        "sample motor C",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    reg.register(cptA)
    reg.register(cptB)
    reg.register(cptC)
    # Pass the device names into the findall method
    result = reg.findall(["sample motor A", "sample motor B"])
    assert cptA in result
    assert cptB in result
    assert cptC not in result


def test_user_readback():
    """Edge case where EpicsMotor.user_readback is named the same as the motor itself."""
    registry = InstrumentRegistry()
    device = EpicsMotor("", name="epics_motor")
    registry.register(device)
    # See if requesting the device.user_readback returns the proper signal
    registry.find("epics_motor_user_readback")



from typing import Union, Sequence

from ophyd import Device, Component, Signal


Detector = Union[Device, Component, Signal, str]


DetectorList = Union[str, Sequence[Detector]]


Motor = Union[Device, Component, Signal, str]



"""A bluesky plan to scan the X-ray energy and capture detector signals.

"""

import logging
import warnings
from typing import Sequence, Union, Mapping
from collections import ChainMap

from bluesky import plans as bp
import numpy as np

from .. import merge_ranges
from .._iconfig import load_config
from ..instrument import registry
from ..constants import edge_energy
from ..typing import DetectorList
from ..preprocessors import baseline_decorator, shutter_suspend_decorator


__all__ = ["energy_scan"]


log = logging.getLogger(__name__)


# @shutter_suspend_decorator()
@baseline_decorator()
def energy_scan(
    energies: Sequence[float],
    exposure: Union[float, Sequence[float]] = 0.1,
    E0: Union[float, str] = 0,
    detectors: DetectorList = "ion_chambers",
    energy_positioners: Sequence = ["energy"],
    time_positioners: Sequence = ["I0_exposure_time"],
    md: Mapping = {},
):
    """Collect a spectrum by scanning X-ray energy.

    For scanning over a pre-defined X-ray absorption edge, try
    :py:func:`~haven.plans.xafs_scan.xafs_scan` instead.

    *exposure* can be either a float, or a sequence of floats. If a
    single value is provided, it will be used for all energies. If a
    sequence is provided, it should be the same length as *energies*
    and the each entry will be used for the corresponding entry in
    *energies*.

    **Usage:**

    The following code will run a scan with 1 eV steps from 13000 to
    13099 eV.

        energies = range(13000, 13100)
        RE(energy_scan(energies, exposure=0.5))

    The preparation of energies is up to the calling function. For
    more sophisticated energy lists, consider using the utility
    functions :py:class:`haven.energy_ranges.ERange` and
    :py:class:`haven.energy_ranges.KRange`:

    .. code-block:: python

        energies = [
          ERange(13000, 13100, E_step=10),
          ERange(13100, 13160, E_step=0.5),
          KRange(13160, k_max=8, k_step=0.2),
        ]
        RE(energy_scan(energies, exposure=0.5))

    The results can also be viewed by passing
    :py:class:`haven.callbacks.live_xafs_plot.LiveXAFSPlot` to the RunEngine:

    .. code-block:: python

        RE(energy_scan(...), LiveXAFSPlot())

    Parameters
    ==========
    energies
      The X-ray energies, in eV, over which to scan.
    exposure
      How long, in seconds, to count at each energy.
    E0
      An edge energy, in eV, or name of the edge of the form
      ``"Ni_L3"``. All energies will be relative to this value.
    detectors
      The detectors to collect X-ray signal from at each energy.
    energy_positioners
      Positioners that will receive the changing energies.
    time_positioners
      Positioners that will receive the exposure time for each scan.
    md
      Additional metadata to pass on the to run engine.

    Yields
    ======
    Bluesky messages to execute the scan.

    """
    # Check that arguments are sensible
    if len(energy_positioners) < 1:
        msg = "Cannot run energy_scan with empty *energy_positioners*."
        log.error(msg)
        raise ValueError(msg)
    # Resolve the detector and positioner list if given by name
    if isinstance(detectors, str):
        detectors = registry.findall(detectors)
    real_detectors = []
    for det in detectors:
        real_detectors.extend(registry.findall(det))
    log.debug(f"Found registered detectors: {real_detectors}")
    energy_positioners = [registry.find(ep) for ep in energy_positioners]
    time_positioners = [registry.find(tp) for tp in time_positioners]
    # Convert an individual exposure time to an array of exposure times
    if not hasattr(exposure, "__iter__"):
        exposure = [exposure] * len(energies)
    # Correct for E0
    if isinstance(E0, str):
        # Look up E0 in the database if e.g. "Ni_K" is given as E0
        E0_str = E0
        E0 = edge_energy(E0)
    else:
        E0_str = None
    energies = np.asarray(energies)
    energies += E0
    # Prepare the positioners list with associated energies and exposures
    scan_args = [(motor, energies) for motor in energy_positioners]
    scan_args += [(motor, exposure) for motor in time_positioners]
    scan_args = [item for items in scan_args for item in items]
    # Do the actual scan
    config = load_config()
    yield from bp.list_scan(
        real_detectors,
        *scan_args,
        md=ChainMap(md, {"edge": E0_str, "E0": E0}, config),
    )



import datetime as dt
from pathlib import Path
import logging

from bluesky.callbacks import LiveFit, best_effort, fitting, mpl_plotting
from bluesky.preprocessors import subs_decorator
from bluesky import plan_stubs as bps
from apstools.plans.alignment import lineup
from matplotlib import pyplot as plt
from lmfit.models import StepModel

from ..instrument.instrument_registry import registry
from .set_energy import set_energy
from .auto_gain import auto_gain
from .beam_properties import knife_scan
from .._iconfig import load_config
from .align_motor import align_pitch2


log = logging.getLogger(__name__)


def calibrate_mono_gap(
    gap_list,
    mono_energies,
    id_energies,
    knife_motor,
    knife_points,
    beamline,
    md={},
    folder="",
):
    """All energies in eV."""
    config = load_config()
    md = dict(beamline=config["beamline"]["name"], **md)
    now = dt.datetime.now()
    now_str = now.strftime("%Y-%m-%d_%H-%M")
    # Find registered ophyd devices
    monochromator = registry.find(name="monochromator")
    It = registry.find(name="It")
    registry.find(name="I0")
    knife_motor = registry.find(any=knife_motor)
    # Gap scans at each gap
    for new_gap in gap_list:
        # Determine the new file name based on gap
        folder = Path(folder).expanduser()
        fname = folder / f"knife_edge_scans_{now_str}_{beamline}_{new_gap}gap.txt"
        fp = fname.resolve()
        fp.parent.mkdir(parents=True, exist_ok=True)
        log.info(f"Saving calibration results to {fp}")
        # Go to the new gap value
        md["gap"] = new_gap
        yield from bps.mv(monochromator.gap, new_gap)
        # Open the file the write
        with open(fp, mode="x") as fd:
            fd.write(
                "mono_energy\tid_energy\tbragg_arcsec\t"
                "pitch2\t\tknife_cen\tknife_dcen\tknife_com\tknife_dcom\n"
            )
        for mono_energy, id_energy in zip(mono_energies, id_energies):
            # Move to the new energy
            yield from set_energy(mono_energy=mono_energy, id_energy=id_energy)
            # Prepare metadata for the scans
            for key in ["pitch2"]:
                if key in md.keys():
                    del md[key]
            md.update(
                {
                    "target_energy": mono_energy,
                    "actual_energy": monochromator.energy.get().user_readback,
                    "bragg_angle": monochromator.bragg.get().user_readback,
                    "undulator_sp": id_energy,
                }
            )
            # Align the mono pitch motor
            bec = best_effort.BestEffortCallback()
            bec.disable_table()
            if mono_energy < 6500:
                pitch_distance = 50
            else:
                pitch_distance = 150
            # Do the cycle twice: align pitch motor, adjust gain, ...
            for idx in [0, 1]:
                yield from subs_decorator(bec)(align_pitch2)(
                    bec=bec, distance=pitch_distance, md=md
                )
                new_pitch2 = bec.peaks["cen"].get("I0_raw_counts", None)
                md["pitch2"] = new_pitch2
                plt.show()
                # Set the pitch2 to the new value (from below accounting for hysteresis)
                yield from bps.mvr(monochromator.pitch2, -200)
                yield from bps.mv(monochromator.pitch2, new_pitch2)
                # Set the ion chamber gain
                yield from auto_gain()
            # Prepare callback for peak states
            knife_scan_ = knife_scan
            peaks = fitting.PeakStats(
                "knife", "It_raw_counts", calc_derivative_and_stats=True
            )
            knife_scan_ = subs_decorator(peaks)(knife_scan_)
            # Callback for fitting with a step function (erf)
            model = StepModel(form="erf")
            fit_cb = LiveFit(
                model,
                It.raw_counts.name,
                {"x": knife_motor.name},
            )
            knife_scan_ = subs_decorator(fit_cb)(knife_scan_)
            # Do the actual knife scan and fitting
            yield from knife_scan_(
                knife_motor=knife_motor, knife_points=knife_points, md=md
            )
            results = peaks
            # Save the energy and center of mass to disk
            getattr(results["stats"], "cen", None)
            getattr(results["derivative_stats"], "cen", None)
            getattr(results["stats"], "com", None)
            getattr(results["derivative_stats"], "com", None)
            log.info(f"Knife center: {results['stats']}")
            with open(fp, mode="a") as fd:
                fd.write(
                    f"{mono_energy}\t{id_energy}\t{md['bragg_angle']}\t"
                    "{new_pitch2}\t{cen}\t{dcen}\t{com}\t{dcom}\n"
                )
            # Plot results
            if hasattr(results, "x_data"):
                plt.figure()
                ax = plt.gca()
                ax.set_title(f"{mono_energy:.1f} eV / {new_gap} µm gap")
                mpl_plotting.plot_peak_stats(results, ax=ax)
                ax.axvline(results["derivative_stats"].com, label="f' center of mass")
                plt.show()



from unittest.mock import MagicMock
from bluesky import plans as bp
from bluesky.callbacks import CallbackBase
from bluesky.simulators import summarize_plan
from ophyd.sim import det, motor, SynAxis, make_fake_device, instantiate_fake_device
import os

from haven import plans, baseline_decorator, baseline_wrapper, run_engine
from haven.preprocessors import shutter_suspend_wrapper, shutter_suspend_decorator
from haven.instrument.aps import EpicsBssDevice, load_aps, ApsMachine


def test_shutter_suspend_wrapper(sim_aps, sim_shutters, sim_registry):
    # Check that the run engine does not have any shutter suspenders
    # Currently this test is fragile since we might add non-shutter
    # suspenders in the future.
    RE = run_engine(connect_databroker=False)
    assert len(RE.suspenders) == 1
    # Check that the shutter suspenders get added
    plan = bp.count([det])
    msgs = list(plan)
    sub_msgs = [m for m in msgs if m[0] == "install_suspender"]
    unsub_msgs = [m for m in msgs if m[0] == "remove_suspender"]
    assert len(sub_msgs) == 0
    assert len(unsub_msgs) == 0
    # Now wrap the plan in the suspend wrapper
    plan = shutter_suspend_wrapper(bp.count([det]))
    msgs = list(plan)
    sub_msgs = [m for m in msgs if m[0] == "install_suspender"]
    unsub_msgs = [m for m in msgs if m[0] == "remove_suspender"]
    assert len(sub_msgs) == 2
    assert len(unsub_msgs) == 2
    # Now wrap the plan in the suspend decorator
    plan = shutter_suspend_decorator()(bp.count)([det])
    msgs = list(plan)
    sub_msgs = [m for m in msgs if m[0] == "install_suspender"]
    unsub_msgs = [m for m in msgs if m[0] == "remove_suspender"]
    assert len(sub_msgs) == 2
    assert len(unsub_msgs) == 2


def test_baseline_wrapper(sim_registry, sim_aps):
    # Create a test device
    motor_baseline = SynAxis(name="baseline_motor", labels={"motors", "baseline"})
    sim_registry.register(motor_baseline)
    # Set up a callback to motor streams generated by runengine
    cb = CallbackBase()
    cb.start = MagicMock()
    cb.descriptor = MagicMock()
    cb.event = MagicMock()
    cb.stop = MagicMock()
    RE = run_engine(connect_databroker=False)
    plan = bp.count([det], num=1)
    plan = baseline_wrapper(plan, devices="baseline")
    RE(plan, cb)
    # Check that the callback has the baseline stream inserted
    assert cb.start.called
    assert cb.descriptor.call_count > 1
    baseline_doc = cb.descriptor.call_args_list[0][0][0]
    primary_doc = cb.descriptor.call_args_list[1][0][0]
    assert baseline_doc["name"] == "baseline"
    assert "baseline_motor" in baseline_doc["data_keys"].keys()


def test_baseline_decorator(sim_registry, sim_aps):
    """Similar to baseline wrapper test, but used as a decorator."""
    # Create the decorated function before anything else
    func = baseline_decorator(devices="motors")(bp.count)
    # Create a test device
    motor_baseline = SynAxis(name="baseline_motor", labels={"motors"})
    sim_registry.register(motor_baseline)
    # Set up a callback to motor streams generated by runengine
    cb = CallbackBase()
    cb.start = MagicMock()
    cb.descriptor = MagicMock()
    cb.event = MagicMock()
    cb.stop = MagicMock()
    RE = run_engine(connect_databroker=False)
    plan = func([det], num=1)
    RE(plan, cb)
    # Check that the callback has the baseline stream inserted
    assert cb.start.called
    assert cb.descriptor.call_count > 1
    baseline_doc = cb.descriptor.call_args_list[0][0][0]
    primary_doc = cb.descriptor.call_args_list[1][0][0]
    assert baseline_doc["name"] == "baseline"
    assert "baseline_motor" in baseline_doc["data_keys"].keys()


def test_metadata(sim_registry, sim_aps, monkeypatch):
    """Similar to baseline wrapper test, but used as a decorator."""
    # Load devices
    bss = instantiate_fake_device(EpicsBssDevice, name="bss", prefix="255id:bss:")
    bss.esaf.esaf_id._readback = "12345"
    bss.esaf.title._readback = "Testing the wetness of water."
    bss.esaf.user_last_names._readback = "Bose, Einstein"
    bss.esaf.user_badges._readback = "287341, 339203, 59208"
    bss.proposal.proposal_id._readback = "25873"
    bss.proposal.title._readback = "Making the world a more interesting place."
    bss.proposal.user_last_names._readback = "Franklin, Watson, Crick"
    bss.proposal.user_badges._readback = "287341, 203884, 59208"
    bss.proposal.mail_in_flag._readback = "1"
    bss.proposal.proprietary_flag._readback = "0"
    bss.esaf.aps_cycle._readback = "2023-2"
    bss.proposal.beamline_name._readback = "255ID-C"
    # bss = FakeBss(name="bss")
    sim_registry.register(bss)
    monkeypatch.setenv("EPICS_HOST_ARCH", "PDP11")
    monkeypatch.setenv("EPICS_CA_MAX_ARRAY_BYTES", "16")
    monkeypatch.setenv("PYEPICS_LIBCA", "/dev/null")
    # Set up a callback to motor streams generated by runengine
    cb = CallbackBase()
    cb.start = MagicMock()
    cb.descriptor = MagicMock()
    cb.event = MagicMock()
    cb.stop = MagicMock()
    RE = run_engine(connect_databroker=False)
    plan = bp.count([det], num=1)
    RE(plan, cb)
    # Check that the callback has the correct metadata
    assert cb.start.called
    assert cb.start.call_count == 1
    start_doc = cb.start.call_args[0][0]
    # Check versions
    assert "versions" in start_doc.keys()
    versions = start_doc["versions"]
    assert "haven" in versions.keys()
    assert versions["haven"] == "0.1.0"
    assert "bluesky" in versions.keys()
    # Check metadata keys
    expected_keys = [
        "uid",
        "time",
        "EPICS_HOST_ARCH",
        "beamline_id",
        "facility_id",
        "xray_source",
        "epics_libca",
        "EPICS_CA_MAX_ARRAY_BYTES",
        "login_id",
        "pid",
        "scan_id",
        "proposal_id",
        "plan_type",
        "plan_name",
        "detectors",
        "hints",
        "parameters",
        "purpose",
        "bss_aps_cycle",
        "bss_beamline_name",
        "esaf_id",
        "esaf_title",
        "esaf_users",
        "esaf_user_badges",
        "mail_in_flag",
        "proposal_title",
        "proprietary_flag",
        "proposal_users",
        "proposal_user_badges",
        "sample_name",
    ]
    missing_keys = set(expected_keys) - set(start_doc.keys())
    assert not missing_keys
    # Check metadata values
    expected_data = {
        "EPICS_HOST_ARCH": "PDP11",
        "beamline_id": "SPC Beamline (sector unknown)",
        "facility_id": "Advanced Photon Source",
        "xray_source": "insertion device",
        "epics_libca": "/dev/null",
        "EPICS_CA_MAX_ARRAY_BYTES": "16",
        "scan_id": 1,
        "plan_type": "generator",
        "plan_name": "count",
        "parameters": "",
        "purpose": "",
        "esaf_id": "12345",
        "esaf_title": "Testing the wetness of water.",
        "esaf_users": "Bose, Einstein",
        "esaf_user_badges": "287341, 339203, 59208",
        "mail_in_flag": "1",
        "proposal_id": "25873",
        "proposal_title": "Making the world a more interesting place.",
        "proposal_users": "Franklin, Watson, Crick",
        "proposal_user_badges": "287341, 203884, 59208",
        "proprietary_flag": "0",
        "sample_name": "",
        "bss_aps_cycle": "2023-2",
        "bss_beamline_name": "255ID-C",
    }
    for key, val in expected_data.items():
        assert start_doc[key] == val, f"{key}: {start_doc[key]}"


# uid: 600852ca-3776-4e7a-ba29-f11786371e55
# time: 1667935604.6147602
# EPICS_HOST_ARCH: linux-x86_64
# beamline_id: APS 9-ID-C USAXS
# epics_libca: >-
#   /home/beams11/USAXS/micromamba/envs/bluesky_2022_3/lib/python3.10/site-packages/epics/clibs/linux64/libca.so
# EPICS_CA_MAX_ARRAY_BYTES: '1280000'
# login_id: usaxs@usaxscontrol.xray.aps.anl.gov
# pid: 1639175
# scan_id: 92
# proposal_id: ''
# versions:
#   apstools: 1.6.8
#   area_detector_handlers: 0.0.10
#   bluesky: 1.10.0
#   databroker: 1.2.5
#   epics_ca: 3.5.0
#   Epics: 3.5.0
#   h5py: 3.7.0
#   matplotlib: 3.6.1
#   numpy: 1.23.3
#   ophyd: 1.7.0
#   pymongo: 4.2.0
#   pyRestTable: 2020.0.6
#   spec2nexus: 2021.2.4
# plan_type: generator
# plan_name: WAXS
# detectors:
#   - waxs_det
# num_points: 1
# num_intervals: 0
# plan_args:
#   detectors:
#     - >-
#       MyPilatusDetector(prefix='usaxs_pilatus2:', name='waxs_det',
#       read_attrs=['hdf1'], configuration_attrs=['cam', 'cam.acquire_period',
#       'cam.acquire_time', 'cam.image_mode', 'cam.manufacturer', 'cam.model',
#       'cam.num_exposures', 'cam.num_images', 'cam.trigger_mode', 'hdf1'])
#   num: 1
# hints:
#   dimensions:
#     - - - time
#       - primary
# full_filename: /share1/USAXS_data/2022-11/usaxs.mac
# filename: usaxs.mac
# line_number: 38
# action: waxsExp
# parameters:
#   - '40'
#   - '120'
#   - '3.77'
#   - Frye_VAC_Perp
# iso8601: '2022-11-08 13:26:41.101412'
# purpose: tuner
# datetime: '2022-11-08 12:48:48.735664'
# sx: 40
# sy: 120
# thickness: 3.77
# title: Frye_VAC_Perp
# bss_aps_cycle: ''
# bss_beamline_name: ''
# esaf_id: ''
# esaf_title: ''
# mail_in_flag: 'OFF'
# principal_user: no users listed
# proposal_title: ''
# proprietary_flag: 'OFF'
# sample_thickness_mm: 3.77
# hdf5_path: /share1/USAXS_data/2022-11/11_08_Frye/11_08_Frye_waxs
# hdf5_file: Frye_VAC_Perp_0009.hdf
# sample_image_name: /share1/USAXS_data/2022-11/11_08_Frye/11_08_Frye_waxs/Frye_VAC_Perp_0009.jpg
# method: areaDetectorAcquire
# area_detector_name: waxs_det



import pytest

from firefly.main_window import FireflyMainWindow
from firefly.xafs_scan import XafsScanDisplay


def test_region_number(qtbot):
    """Does changing the region number affect the UI?"""
    window = FireflyMainWindow()
    qtbot.addWidget(window)
    disp = XafsScanDisplay()
    qtbot.addWidget(disp)
    # Check that the display has the right number of rows to start with
    assert disp.ui.regions_spin_box.value() == 3
    assert hasattr(disp, "regions")
    assert len(disp.regions) == 3
    # Check that regions can be inserted and removed


def test_region(qtbot):
    """Does changing the region ui respond the way it should."""
    window = FireflyMainWindow()
    qtbot.addWidget(window)
    disp = XafsScanDisplay()
    qtbot.addWidget(disp)
    # Does the k-space checkbox enable the k-weight edit line
    region = disp.regions[0]
    region.k_space_checkbox.setChecked(True)
    assert region.k_weight_line_edit.isEnabled() is True


def test_E0_checkbox(qtbot):
    """Does selecting the E0 checkbox adjust the UI properly?"""
    window = FireflyMainWindow()
    qtbot.addWidget(window)
    disp = XafsScanDisplay()
    qtbot.addWidget(disp)
    # K-space checkboxes should be disabled when E0 is unchecked
    disp.ui.use_edge_checkbox.setChecked(False)
    assert not disp.regions[0].k_space_checkbox.isEnabled()
    # K-space checkbox should become re-enabled after E0 is checked
    disp.ui.use_edge_checkbox.setChecked(True)
    assert disp.regions[0].k_space_checkbox.isEnabled()
    # Checked k-space boxes should be unchecked when the E0 is disabled
    disp.regions[0].k_space_checkbox.setChecked(True)
    disp.ui.use_edge_checkbox.setChecked(False)
    disp.regions[0].k_space_checkbox.setChecked(False)
    assert not disp.regions[0].k_space_checkbox.isChecked()



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import (
    PVGroup,
    SubGroup,
    ioc_arg_parser,
    pvproperty,
    run,
    PvpropertyString,
    PvpropertyInteger,
)
from ophyd.tests.fake_motor_ioc import FakeMotorIOC

from haven.simulated_ioc import ResponsiveMotorFields  # , IOC as IOC_


class SimpleGroup(PVGroup):
    """
    An IOC that looks like an undulator.

    E.g. "25IDds:Energy"

    """

    esaf_cycle = pvproperty(value="2023-2", name="esaf:cycle")
    esaf_description = pvproperty(value="", name="esaf:description")
    esaf_enddate = pvproperty(value="", name="esaf:endDate")
    esaf_id = pvproperty(value="12345", name="esaf:id", dtype=PvpropertyString)
    esaf_status = pvproperty(value="", name="esaf:status")
    esaf_number_users_in_pvs = pvproperty(name="esaf:users_in_pvs", value="")
    esaf_number_users_total = pvproperty(name="esaf:users_total", value="")
    esaf_raw = pvproperty(name="esaf:raw", value="")
    esaf_sector = pvproperty(name="esaf:sector", value="")
    esaf_start_date = pvproperty(name="esaf:startDate", value="")
    esaf_title = pvproperty(name="esaf:title", value="Testing the wetness of water.")
    esaf_user_last_names = pvproperty(name="esaf:users", value="Bose, Einstein")
    esaf_user_badges = pvproperty(name="esaf:userBadges", value="287341, 339203, 59208")
    esaf_user1_badge_number = pvproperty(name="esaf:user1:badgeNumber", value="")
    esaf_user1_email = pvproperty(name="esaf:user1:email", value="")
    esaf_user1_first_name = pvproperty(name="esaf:user1:firstName", value="")
    esaf_user1_last_name = pvproperty(name="esaf:user1:lastName", value="")
    esaf_user2_badge_number = pvproperty(name="esaf:user2:badgeNumber", value="")
    esaf_user2_email = pvproperty(name="esaf:user2:email", value="")
    esaf_user2_first_name = pvproperty(name="esaf:user2:firstName", value="")
    esaf_user2_last_name = pvproperty(name="esaf:user2:lastName", value="")
    esaf_user3_badge_number = pvproperty(name="esaf:user3:badgeNumber", value="")
    esaf_user3_email = pvproperty(name="esaf:user3:email", value="")
    esaf_user3_first_name = pvproperty(name="esaf:user3:firstName", value="")
    esaf_user3_last_name = pvproperty(name="esaf:user3:lastName", value="")
    esaf_user4_badge_number = pvproperty(name="esaf:user4:badgeNumber", value="")
    esaf_user4_email = pvproperty(name="esaf:user4:email", value="")
    esaf_user4_first_name = pvproperty(name="esaf:user4:firstName", value="")
    esaf_user4_last_name = pvproperty(name="esaf:user4:lastName", value="")
    esaf_user5_badge_number = pvproperty(name="esaf:user5:badgeNumber", value="")
    esaf_user5_email = pvproperty(name="esaf:user5:email", value="")
    esaf_user5_first_name = pvproperty(name="esaf:user5:firstName", value="")
    esaf_user5_last_name = pvproperty(name="esaf:user5:lastName", value="")
    esaf_user6_badge_number = pvproperty(name="esaf:user6:badgeNumber", value="")
    esaf_user6_email = pvproperty(name="esaf:user6:email", value="")
    esaf_user6_first_name = pvproperty(name="esaf:user6:firstName", value="")
    esaf_user6_last_name = pvproperty(name="esaf:user6:lastName", value="")
    esaf_user7_badge_number = pvproperty(name="esaf:user7:badgeNumber", value="")
    esaf_user7_email = pvproperty(name="esaf:user7:email", value="")
    esaf_user7_first_name = pvproperty(name="esaf:user7:firstName", value="")
    esaf_user7_last_name = pvproperty(name="esaf:user7:lastName", value="")
    esaf_user8_badge_number = pvproperty(name="esaf:user8:badgeNumber", value="")
    esaf_user8_email = pvproperty(name="esaf:user8:email", value="")
    esaf_user8_first_name = pvproperty(name="esaf:user8:firstName", value="")
    esaf_user8_last_name = pvproperty(name="esaf:user8:lastName", value="")
    esaf_user9_badge_number = pvproperty(name="esaf:user9:badgeNumber", value="")
    esaf_user9_email = pvproperty(name="esaf:user9:email", value="")
    esaf_user9_first_name = pvproperty(name="esaf:user9:firstName", value="")
    esaf_user9_last_name = pvproperty(name="esaf:user9:lastName", value="")
    proposal_beamline_name = pvproperty(name="proposal:beamline", value="99ID-C")
    proposal_end_date = pvproperty(name="proposal:endDate", value="")
    proposal_mail_in_flag = pvproperty(
        name="proposal:mailInFlag", value=1, dtype=PvpropertyInteger
    )
    proposal_number_users_in_pvs = pvproperty(name="proposal:users_in_pvs", value="")
    proposal_number_users_total = pvproperty(name="proposal:users_total", value="")
    proposal_id = pvproperty(name="proposal:id", value="25873", dtype=PvpropertyString)
    proposal_proprietary_flag = pvproperty(
        name="proposal:proprietaryFlag", value=0, dtype=PvpropertyInteger
    )
    proposal_raw = pvproperty(name="proposal:raw", value="")
    proposal_start_date = pvproperty(name="proposal:startDate", value="")
    proposal_submitted_date = pvproperty(name="proposal:submittedDate", value="")
    proposal_title = pvproperty(
        name="proposal:title", value="Making the world a more interesting place."
    )
    proposal_user_badges = pvproperty(
        name="proposal:userBadges", value="287341, 203884, 59208"
    )
    proposal_user_last_names = pvproperty(
        name="proposal:users", value="Franklin, Watson, Crick"
    )
    proposal_user1_badge_number = pvproperty(
        name="proposal:user1:badgeNumber", value=""
    )
    proposal_user1_email = pvproperty(name="proposal:user1:email", value="")
    proposal_user1_first_name = pvproperty(name="proposal:user1:firstName", value="")
    proposal_user1_institution = pvproperty(name="proposal:user1:institution", value="")
    proposal_user1_institution_id = pvproperty(name="proposal:user1:instId", value="")
    proposal_user1_last_name = pvproperty(name="proposal:user1:lastName", value="")
    proposal_user1_pi_flag = pvproperty(name="proposal:user1:piFlag", value="")
    proposal_user1_user_id = pvproperty(name="proposal:user1:userId", value="")
    proposal_user2_badge_number = pvproperty(
        name="proposal:user2:badgeNumber", value=""
    )
    proposal_user2_email = pvproperty(name="proposal:user2:email", value="")
    proposal_user2_first_name = pvproperty(name="proposal:user2:firstName", value="")
    proposal_user2_institution = pvproperty(name="proposal:user2:institution", value="")
    proposal_user2_institution_id = pvproperty(name="proposal:user2:instId", value="")
    proposal_user2_last_name = pvproperty(name="proposal:user2:lastName", value="")
    proposal_user2_pi_flag = pvproperty(name="proposal:user2:piFlag", value="")
    proposal_user2_user_id = pvproperty(name="proposal:user2:userId", value="")
    proposal_user3_badge_number = pvproperty(
        name="proposal:user3:badgeNumber", value=""
    )
    proposal_user3_email = pvproperty(name="proposal:user3:email", value="")
    proposal_user3_first_name = pvproperty(name="proposal:user3:firstName", value="")
    proposal_user3_institution = pvproperty(name="proposal:user3:institution", value="")
    proposal_user3_institution_id = pvproperty(name="proposal:user3:instId", value="")
    proposal_user3_last_name = pvproperty(name="proposal:user3:lastName", value="")
    proposal_user3_pi_flag = pvproperty(name="proposal:user3:piFlag", value="")
    proposal_user3_user_id = pvproperty(name="proposal:user3:userId", value="")
    proposal_user4_badge_number = pvproperty(
        name="proposal:user4:badgeNumber", value=""
    )
    proposal_user4_email = pvproperty(name="proposal:user4:email", value="")
    proposal_user4_first_name = pvproperty(name="proposal:user4:firstName", value="")
    proposal_user4_institution = pvproperty(name="proposal:user4:institution", value="")
    proposal_user4_institution_id = pvproperty(name="proposal:user4:instId", value="")
    proposal_user4_last_name = pvproperty(name="proposal:user4:lastName", value="")
    proposal_user4_pi_flag = pvproperty(name="proposal:user4:piFlag", value="")
    proposal_user4_user_id = pvproperty(name="proposal:user4:userId", value="")
    proposal_user5_badge_number = pvproperty(
        name="proposal:user5:badgeNumber", value=""
    )
    proposal_user5_email = pvproperty(name="proposal:user5:email", value="")
    proposal_user5_first_name = pvproperty(name="proposal:user5:firstName", value="")
    proposal_user5_institution = pvproperty(name="proposal:user5:institution", value="")
    proposal_user5_institution_id = pvproperty(name="proposal:user5:instId", value="")
    proposal_user5_last_name = pvproperty(name="proposal:user5:lastName", value="")
    proposal_user5_pi_flag = pvproperty(name="proposal:user5:piFlag", value="")
    proposal_user5_user_id = pvproperty(name="proposal:user5:userId", value="")
    proposal_user6_badge_number = pvproperty(
        name="proposal:user6:badgeNumber", value=""
    )
    proposal_user6_email = pvproperty(name="proposal:user6:email", value="")
    proposal_user6_first_name = pvproperty(name="proposal:user6:firstName", value="")
    proposal_user6_institution = pvproperty(name="proposal:user6:institution", value="")
    proposal_user6_institution_id = pvproperty(name="proposal:user6:instId", value="")
    proposal_user6_last_name = pvproperty(name="proposal:user6:lastName", value="")
    proposal_user6_pi_flag = pvproperty(name="proposal:user6:piFlag", value="")
    proposal_user6_user_id = pvproperty(name="proposal:user6:userId", value="")
    proposal_user7_badge_number = pvproperty(
        name="proposal:user7:badgeNumber", value=""
    )
    proposal_user7_email = pvproperty(name="proposal:user7:email", value="")
    proposal_user7_first_name = pvproperty(name="proposal:user7:firstName", value="")
    proposal_user7_institution = pvproperty(name="proposal:user7:institution", value="")
    proposal_user7_institution_id = pvproperty(name="proposal:user7:instId", value="")
    proposal_user7_last_name = pvproperty(name="proposal:user7:lastName", value="")
    proposal_user7_pi_flag = pvproperty(name="proposal:user7:piFlag", value="")
    proposal_user7_user_id = pvproperty(name="proposal:user7:userId", value="")
    proposal_user8_badge_number = pvproperty(
        name="proposal:user8:badgeNumber", value=""
    )
    proposal_user8_email = pvproperty(name="proposal:user8:email", value="")
    proposal_user8_first_name = pvproperty(name="proposal:user8:firstName", value="")
    proposal_user8_institution = pvproperty(name="proposal:user8:institution", value="")
    proposal_user8_institution_id = pvproperty(name="proposal:user8:instId", value="")
    proposal_user8_last_name = pvproperty(name="proposal:user8:lastName", value="")
    proposal_user8_pi_flag = pvproperty(name="proposal:user8:piFlag", value="")
    proposal_user8_user_id = pvproperty(name="proposal:user8:userId", value="")
    proposal_user9_badge_number = pvproperty(
        name="proposal:user9:badgeNumber", value=""
    )
    proposal_user9_email = pvproperty(name="proposal:user9:email", value="")
    proposal_user9_first_name = pvproperty(name="proposal:user9:firstName", value="")
    proposal_user9_institution = pvproperty(name="proposal:user9:institution", value="")
    proposal_user9_institution_id = pvproperty(name="proposal:user9:instId", value="")
    proposal_user9_last_name = pvproperty(name="proposal:user9:lastName", value="")
    proposal_user9_pi_flag = pvproperty(name="proposal:user9:piFlag", value="")
    proposal_user9_user_id = pvproperty(name="proposal:user9:userId", value="")
    ioc_host = pvproperty(name="ioc_host", value="")
    ioc_user = pvproperty(name="ioc_user", value="")
    status_msg = pvproperty(name="status", value="")

    default_prefix = "99id:bss:"


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idc:bss:", desc="haven.tests.ioc_apsbss test IOC"
    )
    ioc = SimpleGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



__all__ = ["energy_scan"]

__version__ = "0.1.0"

#  Top-level imports
from .catalog import load_catalog, load_result, load_data, tiled_client  # noqa: F401
from .energy_ranges import ERange, KRange, merge_ranges  # noqa: F401
from .plans.energy_scan import energy_scan  # noqa: F401
from .plans.align_slits import align_slits  # noqa: F401
from .plans.beam_properties import knife_scan, fit_step  # noqa: F401
from .plans.xafs_scan import xafs_scan  # noqa: F401
from .plans.beam_properties import knife_scan  # noqa: F401
from .plans.auto_gain import auto_gain, AutoGainCallback  # noqa:F401
from .plans.mono_gap_calibration import calibrate_mono_gap  # noqa: F401
from .plans.mono_ID_calibration import mono_ID_calibration  # noqa: F401
from .plans.set_energy import set_energy  # noqa: F401
from .plans.align_motor import align_motor, align_pitch2  # noqa: F401
from .plans.shutters import open_shutters, close_shutters  # noqa: F401
from .plans.record_dark_current import record_dark_current  # noqa: F401
from .run_engine import run_engine  # noqa: F401
from ._iconfig import load_config  # noqa: F401
from .preprocessors import (
    baseline_decorator,
    baseline_wrapper,
    shutter_suspend_wrapper,
    shutter_suspend_decorator,
)  # noqa: F401
from .constants import edge_energy

from .motor_position import (  # noqa: F401
    save_motor_position,
    list_motor_positions,
    get_motor_position,
    recall_motor_position,
    list_current_motor_positions,
)
from .instrument import (  # noqa: F401
    ion_chamber,
    IonChamber,
    InstrumentRegistry,
    registry,
    Monochromator,
)
from .instrument.load_instrument import load_instrument  # noqa: F401
from .instrument.fluorescence_detector import (
    load_dxp_detector,
    load_fluorescence_detectors,
)
from .instrument.motor import HavenMotor  # noqa: F401

from .xdi_writer import XDIWriter  # noqa: F401
from .progress_bar import ProgressBar  # noqa: F401



from haven.instrument.load_instrument import load_simulated_devices


def test_load_simulated_devices(sim_registry):
    load_simulated_devices()
    # Check motors
    motor = sim_registry.find(name="sim_motor")
    # Check detectors
    detector = sim_registry.find(name="sim_detector")



import logging

from qtpy import QtWidgets, QtCore
from bluesky_queueserver_api import BPlan
from haven import registry, load_config
from xraydb.xraydb import XrayDB

from firefly import display


log = logging.getLogger(__name__)


class EnergyDisplay(display.FireflyDisplay):
    caqtdm_mono_ui_file = "/net/s25data/xorApps/ui/DCMControlCenter.ui"
    caqtdm_id_ui_file = (
        "/net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/IDControl.ui"
    )
    min_energy = 4000
    max_energy = 33000
    stylesheet_danger = (
        "background: rgb(220, 53, 69); color: white; border-color: rgb(220, 53, 69)"
    )
    stylesheet_normal = ""

    def __init__(self, args=None, macros={}, **kwargs):
        # Load X-ray database for calculating edge energies
        self.xraydb = XrayDB()
        super().__init__(args=args, macros=macros, **kwargs)

    def launch_mono_caqtdm(self):
        config = load_config()
        prefix = config["monochromator"]["ioc"] + ":"
        mono = registry.find(name="monochromator")
        ID = registry.find(name="undulator")
        caqtdm_macros = {
            "P": prefix,
            "MONO": config["monochromator"]["ioc_branch"],
            "BRAGG": mono.bragg.prefix.replace(prefix, ""),
            "GAP": mono.gap.prefix.replace(prefix, ""),
            "ENERGY": mono.energy.prefix.replace(prefix, ""),
            "OFFSET": mono.offset.prefix.replace(prefix, ""),
            "IDENERGY": ID.energy.pvname,
        }
        self.launch_caqtdm(macros=caqtdm_macros, ui_file=self.caqtdm_mono_ui_file)

    def launch_id_caqtdm(self):
        """Launch the pre-built caQtDM UI file for the ID."""
        config = load_config()
        prefix = config["undulator"]["ioc"]
        # Strip leading "ID" from the mono IOC since caQtDM adds it
        prefix = prefix.strip("ID")
        caqtdm_macros = {
            # No idea what "M", and "D" do, they're not in the UI
            # file.
            "ID": prefix,
            "M": 2,
            "D": 2,
        }
        self.launch_caqtdm(macros=caqtdm_macros, ui_file=self.caqtdm_id_ui_file)

    def set_energy(self, *args, **kwargs):
        energy = float(self.ui.target_energy_lineedit.text())
        log.info(f"Setting new energy: {energy}")
        # Build the queue item
        item = BPlan("set_energy", energy=energy)
        # Submit the item to the queueserver
        from firefly.application import FireflyApplication

        app = FireflyApplication.instance()
        app.add_queue_item(item)

    def customize_ui(self):
        self.ui.mono_caqtdm_button.clicked.connect(self.launch_mono_caqtdm)
        self.ui.id_caqtdm_button.clicked.connect(self.launch_id_caqtdm)
        self.ui.set_energy_button.clicked.connect(self.set_energy)
        # Set up the combo box with X-ray energies
        combo_box = self.ui.edge_combo_box
        ltab = self.xraydb.tables["xray_levels"]
        edges = self.xraydb.query(ltab)
        edges = edges.filter(
            ltab.c.absorption_edge < self.max_energy,
            ltab.c.absorption_edge > self.min_energy,
        )
        items = [
            f"{r.element} {r.iupac_symbol} ({int(r.absorption_edge)} eV)"
            for r in edges.all()
        ]
        combo_box.addItems(["Select edge…", *items])
        combo_box.activated.connect(self.select_edge)

    @QtCore.Slot(int)
    def select_edge(self, index):
        if index == 0:
            # The placeholder text was selected
            return
        # Parse the combo box text to get the selected edge
        combo_box = self.ui.edge_combo_box
        text = combo_box.itemText(index)
        elem, edge = text.replace(" ", "_").split("_")[:2]
        # Determine which energy was selected
        edge_info = self.xraydb.xray_edge(element=elem, edge=edge)
        if edge_info is None:
            # Edge is not recognized, so provide feedback
            combo_box.setStyleSheet(self.stylesheet_danger)
        else:
            # Set the text field to the selected edge's energy
            energy, fyield, edge_jump = edge_info
            self.ui.target_energy_lineedit.setText(f"{energy:.3f}")
            combo_box.setStyleSheet(self.stylesheet_normal)

    def ui_filename(self):
        return "energy.ui"



import qtawesome as qta
from pydm import PyDMChannel
from qtpy.QtCore import Signal

from firefly import display


class XRFROIDisplay(display.FireflyDisplay):
    enabled_background = "rgb(212, 237, 218)"  # Pale green
    selected_background = "rgb(204, 229, 255)"  # Pale blue

    # Signals
    selected = Signal(bool)

    def ui_filename(self):
        return "xrf_roi.ui"

    def customize_ui(self):
        self.ui.set_roi_button.setIcon(qta.icon("fa5s.chart-line"))
        self.ui.enabled_checkbox.toggled.connect(self.set_backgrounds)
        self.ui.set_roi_button.toggled.connect(self.set_backgrounds)
        self.ui.enabled_checkbox.toggled.connect(self.enable_roi)
        self.ui.set_roi_button.toggled.connect(self.selected)

    def set_backgrounds(self):
        is_selected = self.ui.set_roi_button.isChecked()
        if is_selected:
            self.setStyleSheet(f"background: {self.selected_background}")
        else:
            self.setStyleSheet("")

    def enable_roi(self, is_enabled):
        if not is_enabled:
            # Unselect this channel so we don't get locked out
            self.ui.set_roi_button.setChecked(False)



import time

import epics

from haven.instrument.energy_positioner import EnergyPositioner


def test_pseudo_to_real_positioner(ioc_mono, ioc_undulator):
    positioner = EnergyPositioner(
        name="energy",
        mono_pv=ioc_mono.pvs["energy"],
        id_prefix=ioc_undulator.prefix.strip(":"),
        id_tracking_pv=ioc_mono.pvs["id_tracking"],
        id_offset_pv=ioc_mono.pvs["id_offset"],
    )
    positioner.mono_energy.wait_for_connection()
    positioner.id_energy.wait_for_connection()
    positioner.energy.set(10000, timeout=5.0)
    assert positioner.get(use_monitor=False).mono_energy.user_setpoint == 10000
    positioner.id_offset.set(230)
    time.sleep(0.1)
    # Move the energy positioner
    positioner.energy.set(5000)
    time.sleep(0.1)  # Caproto breaks pseudopositioner status
    # Check that the mono and ID are both moved
    assert positioner.get(use_monitor=False).mono_energy.user_setpoint == 5000
    expected_id_energy = 5.0 + positioner.id_offset.get(use_monitor=False) / 1000
    assert positioner.get(use_monitor=False).id_energy.setpoint == expected_id_energy


def test_real_to_pseudo_positioner(ioc_mono, ioc_undulator):
    positioner = EnergyPositioner(
        name="energy",
        mono_pv=ioc_mono.pvs["energy"],
        id_prefix=ioc_undulator.prefix.strip(":"),
        id_tracking_pv=ioc_mono.pvs["id_tracking"],
        id_offset_pv=ioc_mono.pvs["id_offset"],
    )
    positioner.wait_for_connection(timeout=10.0)
    # Move the mono energy positioner
    epics.caput(ioc_mono.pvs["energy"], 5000.0)
    time.sleep(0.1)  # Caproto breaks pseudopositioner status
    assert epics.caget(ioc_mono.pvs["energy"], use_monitor=False) == 5000.0
    # assert epics.caget("mono_ioc:Energy.RBV") == 5000.0
    # Check that the pseudo single is updated
    assert positioner.energy.get(use_monitor=False).readback == 5000.0



import time
import logging
import subprocess
from pathlib import Path
from typing import Sequence, Optional
import json
from contextlib import contextmanager
from functools import partial
from collections import defaultdict
import gc
from enum import IntEnum

from qtpy import uic
from qtpy.QtCore import Qt, Signal, QObject, QThread
from qtpy.QtWidgets import QWidget
import qtawesome as qta
import pyqtgraph
import pydm
from pydm.widgets import PyDMEmbeddedDisplay, PyDMChannel
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import TABLEAU_COLORS

import haven
from firefly import display, FireflyApplication

import sys

np.set_printoptions(threshold=sys.maxsize)


log = logging.getLogger(__name__)


pyqtgraph.setConfigOption("imageAxisOrder", "row-major")

colors = list(TABLEAU_COLORS.values())


class AcquireStates(IntEnum):
    DONE = 0
    ACQUIRING = 1


class ROIRegion(pyqtgraph.LinearRegionItem):
    mca_num: int
    roi_num: int

    lo_channel: PyDMChannel
    hi_channel: PyDMChannel

    region_upper_changed = Signal(int)
    region_lower_changed = Signal(int)

    _last_upper: int = None
    _last_lower: int = None

    def __init__(self, address: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.address = address
        # Set up channels to the IOC
        self.hi_channel = PyDMChannel(
            address=f"{address}.hi_chan._write_pv",
            value_slot=self.set_region_upper,
            value_signal=self.region_upper_changed,
        )
        self.hi_channel.connect()
        self.lo_channel = PyDMChannel(
            address=f"{address}.lo_chan._write_pv",
            value_slot=self.set_region_lower,
            value_signal=self.region_lower_changed,
        )
        self.lo_channel.connect()
        self.sigRegionChangeFinished.connect(self.handle_region_change)
        # Set initial display state
        self.setVisible(False)

    def handle_region_change(self):
        lower, upper = self.getRegion()
        lower = round(lower)
        upper = round(upper)
        if lower != self._last_lower:
            log.debug(f"Changing lower from {self._last_lower} to {lower}")
            if self._last_lower is not None:
                self.region_lower_changed.emit(lower)
            self._last_lower = lower
        if upper != self._last_upper:
            log.debug(f"Changing upper from {self._last_upper} to {upper}")
            if self._last_upper is not None:
                self.region_upper_changed.emit(upper)
            self._last_upper = upper

    def set_region_lower(self, new_lower):
        """Set the upper value of the highlighted region."""
        log.debug(f"Setting new region lower bound: {new_lower}")
        old_region = self.getRegion()
        new_region = (new_lower, old_region[1])
        if new_region != old_region:
            self.setRegion(new_region)
        self._last_lower = new_lower

    def set_region_upper(self, new_upper):
        """Set the upper value of the highlighted region."""
        log.debug(f"Setting new region upper bound: {new_upper}")
        old_region = self.getRegion()
        new_region = (old_region[0], new_upper)
        if new_region != old_region:
            self.setRegion(new_region)
        self._last_upper = new_upper


class XRF1DPlotItem(pyqtgraph.PlotItem):
    """The axes used for plotting."""

    hover_coords_changed = Signal(str)

    def hoverEvent(self, event):
        super().hoverEvent(event)
        if event.isExit():
            self.hover_coords_changed.emit("NaN")
            return
        # Get data coordinates from event
        pos = event.scenePos()
        data_pos = self.vb.mapSceneToView(pos)
        pos_str = f"({data_pos.x():.3f}, {data_pos.y():.3f})"
        self.hover_coords_changed.emit(pos_str)


class XRF1DPlotWidget(pyqtgraph.PlotWidget):
    """The inner widget containing just the plot."""

    def __init__(self, parent=None, background="default", plotItem=None, **kargs):
        plot_item = XRF1DPlotItem(**kargs)
        super().__init__(parent=parent, background=background, plotItem=plot_item)


class XRFPlotWidget(QWidget):
    """The outer widget, containing the plot and related controls."""

    ui_dir = Path(__file__).parent
    _data_items: defaultdict
    _selected_spectrum: int = None
    _region_items: dict
    device_name: str = ""
    target_mca: int = None

    # Signals
    plot_changed = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data_items = defaultdict(lambda: None)
        self._region_items = {}
        self.ui = uic.loadUi(self.ui_dir / "xrf_plot.ui", self)
        # Create plotting items
        plot_item = self.ui.plot_widget.getPlotItem()
        plot_item.addLegend()
        plot_item.hover_coords_changed.connect(self.ui.coords_label.setText)

    def region(self, mca_num, roi_num):
        key = (mca_num, roi_num)
        plot_item = self.ui.plot_widget.getPlotItem()
        # Create a new region item if necessary
        if key not in self._region_items.keys():
            address = f"oph://{self.device_name}.mcas.mca{mca_num}.rois.roi{roi_num}"
            color = self.region_color(mca_num=mca_num, roi_num=roi_num)
            region = ROIRegion(
                address=address,
                brush=f"{color}10",
                hoverBrush=f"{color}20",
                pen=f"{color}50",
            )
            plot_item.addItem(region)
            self._region_items[key] = region
        # Return the region item
        return self._region_items[key]

    def handle_region_change(self, mca_num, roi_num, region):
        lower, upper = region.getRegion()
        lower = round(lower)
        upper = round(upper)
        self.lo_signal(mca_num=mca_num, roi_num=roi_num).emit(lower)
        self.hi_signal(mca_num=mca_num, roi_num=roi_num).emit(upper)

    def update_spectrum(self, mca_num, spectrum):
        """Plot the spectrum associated with the given MCA index."""
        # Create the plot item itself if necessary
        log.debug(f"New spectrum ({mca_num}): {repr(spectrum)}")
        show_spectrum = self.target_mca is None or mca_num == self.target_mca
        row, col = (0, 0)
        plot_item = self.ui.plot_widget.getPlotItem()
        # Get rid of the previous plots
        if (existing_item := self._data_items[mca_num]) is not None:
            plot_item.removeItem(existing_item)
        # Plot the spectrum
        if show_spectrum:
            xdata = np.arange(len(spectrum))
            color = self.spectrum_color(mca_num)
            self._data_items[mca_num] = plot_item.plot(
                xdata, spectrum, name=mca_num, pen=color
            )
            # Add region markers
            self.plot_changed.emit()

    def spectrum_color(self, mca_num):
        return colors[(mca_num - 1) % len(colors)]

    def highlight_spectrum(self, mca_num, roi_num, hovered):
        """Highlight a spectrum and lowlight the rest.

        *mca_num* is the number (1-indexed) of the spectrum plotted.

        If *hovered* is true, then set the highlights, otherwise clear
        them.

        """
        raise NotImplementedError

    def show_region(self, show: bool, mca_num: int = 1, roi_num: int = 0):
        # Hide all the other regions
        for (mca, roi), region in self._region_items.items():
            region.setVisible(False)
        # Show/hide the specific region requested
        if show:
            log.debug(f"Showing region mca={mca_num} roi={roi_num}")
            self.region(mca_num=mca_num, roi_num=roi_num).setVisible(show)


class MCAPlotWidget(XRFPlotWidget):
    def region_color(self, mca_num, roi_num):
        return self.spectrum_color(mca_num)

    def highlight_spectrum(self, mca_num, roi_num, hovered):
        """Highlight a spectrum and lowlight the rest.

        *mca_num* is the number (1-indexed) of the spectrum plotted.

        If *hovered* is true, then set the highlights, otherwise clear
        them.

        """
        # Get the actual line plot
        solid, semisolid, transparent = (1.0, 0.55, 0.15)
        for key, data_item in self._data_items.items():
            if data_item is None:
                continue
            elif hovered:
                # Highlight this specific spectrum
                if key == mca_num:
                    opacity = solid
                elif key == self._selected_spectrum:
                    opacity = semisolid
                else:
                    opacity = transparent
                data_item.setOpacity(opacity)
            elif self._selected_spectrum is not None:
                # Highlight the spectrum that was previously selected
                log.debug(
                    f"Reverting to previously selected spectrum: {self._selected_spectrum}"
                )
                is_dimmed = key != self._selected_spectrum
                if is_dimmed:
                    data_item.setOpacity(transparent)
                else:
                    data_item.setOpacity(solid)
            else:
                # Just make them all solid again
                data_item.setOpacity(solid)
        # Hide or unhide the selecting region
        is_selected = self._selected_spectrum is not None
        if hovered:
            self.show_region(show=True, mca_num=mca_num, roi_num=roi_num)
        elif is_selected:
            self.show_region(
                show=True, mca_num=self._selected_spectrum, roi_num=roi_num
            )
        else:
            self.show_region(show=False, mca_num=mca_num, roi_num=roi_num)

    def select_spectrum(self, mca_num, roi_num, is_selected):
        """Select an active spectrum to modify."""
        log.debug(f"Selecting spectrum {mca_num}")
        self._selected_spectrum = mca_num if is_selected else None
        self.highlight_spectrum(mca_num=mca_num, roi_num=roi_num, hovered=is_selected)


class ROIPlotWidget(XRFPlotWidget):
    def region_color(self, mca_num, roi_num):
        return colors[(roi_num) % len(colors)]

    def select_roi(self, mca_num, roi_num, is_selected):
        """Select an active spectrum to modify."""
        log.debug(f"Selecting ROI {roi_num}")
        self._selected_roi = roi_num if is_selected else None
        self.highlight_spectrum(mca_num=mca_num, roi_num=roi_num, hovered=is_selected)

    def highlight_spectrum(self, mca_num, roi_num, hovered):
        """Highlight a spectrum and lowlight the rest.

        *mca_num* is the number (1-indexed) of the spectrum plotted.

        If *hovered* is true, then set the highlights, otherwise clear
        them.

        """
        # Hide or unhide the selecting region
        is_selected = self._selected_roi is not None
        if hovered:
            self.show_region(show=True, mca_num=mca_num, roi_num=roi_num)
        elif is_selected:
            self.show_region(show=True, mca_num=mca_num, roi_num=self._selected_roi)
        else:
            self.show_region(show=False, mca_num=mca_num, roi_num=roi_num)


class ROIEmbeddedDisplay(PyDMEmbeddedDisplay):
    # Signals
    selected = Signal(bool)
    hovered = Signal(bool)

    def open_file(self, **kwargs):
        widget = super().open_file(**kwargs)
        # Connect signals if necessary
        if widget is not None:
            widget.selected.connect(self.selected)
        return widget

    def enterEvent(self, event=None):
        self.hovered.emit(True)

    def leaveEvent(self, event=None):
        self.hovered.emit(False)


class XrfTriggers(QObject):
    accumulate: bool = False

    # Signals
    start_all = Signal(int)  # is_started
    start_erase = Signal(int)  # (is_started)

    def __init__(self, device, *args, **kwargs):
        self.device = device
        super().__init__(*args, **kwargs)
        self.setup_trigger_channels()

    def set_accumulate(self, accumulate):
        self.accumulate = accumulate

    def setup_trigger_channels(self):
        # Set up a channel for starting detector acquisition
        device = self.device
        self.start_channel = PyDMChannel(
            address=f"oph://{device.name}.start_all",
            value_signal=self.start_all,
        )
        self.start_channel.connect()
        self.start_erase_channel = PyDMChannel(
            address=f"oph://{device.name}.erase_start",
            value_signal=self.start_erase,
        )
        self.start_erase_channel.connect()
        # This one gets (dis)connected in response to the continuous button
        self.acquiring_channel = PyDMChannel(
            address=f"oph://{device.name}.acquiring",
            value_slot=self.trigger_next,
        )

    def trigger_continuously(self, is_started):
        if is_started:
            self.acquiring_channel.connect()
            # Trigger once to start the process
            self.trigger_once(is_started=is_started)
        else:
            self.acquiring_channel.disconnect()

    def trigger_next(self, acquire_state):
        """Check if acquiring is done and start the next frame.

        Mostly used for continuous acquisition.

        """
        is_started = acquire_state == AcquireStates.DONE
        self.trigger_once(is_started=is_started)

    def trigger_once(self, is_started):
        log.debug("Triggering once")
        if self.accumulate:
            start_signal = self.start_all
        else:
            start_signal = self.start_erase
        start_signal.emit(1)


class XRFDetectorDisplay(display.FireflyDisplay):
    roi_displays: Sequence = []
    mca_displays: Sequence = []
    _spectrum_channels: Sequence
    _selected_mca: int = None
    _mca_lower_receiver = None
    _mca_upper_receiver = None

    # Signals
    spectrum_changed = Signal(int, object)  # (MCA index, spectrum)
    mca_row_hovered = Signal(int, int, bool)  # (MCA num, roi_num, entered)
    roi_row_hovered = Signal(int, int, bool)  # (MCA num, roi_num, entered)

    def ui_filename(self):
        return "xrf_detector.ui"

    def customize_ui(self):
        device = self.device
        self.setWindowTitle(device.name)
        self.ui.mca_plot_widget.device_name = self.device.name
        self.ui.roi_plot_widget.device_name = self.device.name
        # Set ROI and element selection comboboxes
        self.ui.mca_combobox.currentIndexChanged.connect(self.draw_roi_widgets)
        self.ui.roi_combobox.currentIndexChanged.connect(self.draw_mca_widgets)
        elements = [str(i) for i in range(1, device.num_elements + 1)]
        self.ui.mca_combobox.addItems(elements)
        rois = [str(i) for i in range(device.num_rois)]
        self.ui.roi_combobox.addItems(rois)
        # Controls for increment/decrement ROI/MCA combobox
        self.ui.mca_up_button.setIcon(qta.icon("fa5s.arrow-right"))
        self.ui.mca_down_button.setIcon(qta.icon("fa5s.arrow-left"))
        self.ui.roi_up_button.setIcon(qta.icon("fa5s.arrow-right"))
        self.ui.roi_down_button.setIcon(qta.icon("fa5s.arrow-left"))
        self.ui.mca_up_button.clicked.connect(
            partial(self.increment_combobox, combobox=self.ui.mca_combobox, step=1)
        )
        self.ui.mca_down_button.clicked.connect(
            partial(self.increment_combobox, combobox=self.ui.mca_combobox, step=-1)
        )
        self.ui.roi_up_button.clicked.connect(
            partial(self.increment_combobox, combobox=self.ui.roi_combobox, step=1)
        )
        self.ui.roi_down_button.clicked.connect(
            partial(self.increment_combobox, combobox=self.ui.roi_combobox, step=-1)
        )
        # Button for starting/stopping the detector
        triggers = XrfTriggers(device=self.device)
        self.ui.continuous_button.setIcon(qta.icon("fa5s.play"))
        self.ui.erase_button.setIcon(qta.icon("fa5s.eraser"))
        self.ui.continuous_button.toggled.connect(triggers.trigger_continuously)
        self.ui.oneshot_button.clicked.connect(triggers.trigger_once)
        self.ui.accumulate_checkbox.toggled.connect(triggers.set_accumulate)
        self.triggers = triggers
        # Run the worker for starting/stopping the detector in a separate thread
        thread = QThread()
        triggers.moveToThread(thread)
        thread.finished.connect(triggers.deleteLater)
        thread.start()
        self.triggers_thread = thread
        # Buttons for modifying all ROI settings
        self.ui.mca_copyall_button.setIcon(qta.icon("fa5.clone"))
        self.ui.mca_copyall_button.clicked.connect(self.copy_selected_mca)
        self.ui.roi_copyall_button.setIcon(qta.icon("fa5.clone"))
        self.ui.roi_copyall_button.clicked.connect(self.copy_selected_roi)
        self.ui.mca_enableall_checkbox.setCheckState(Qt.PartiallyChecked)
        self.ui.mca_enableall_checkbox.stateChanged.connect(self.enable_mca_checkboxes)
        self.ui.roi_enableall_checkbox.setCheckState(Qt.PartiallyChecked)
        self.ui.roi_enableall_checkbox.stateChanged.connect(self.enable_roi_checkboxes)
        # Connect signals for when spectra change
        self.spectrum_changed.connect(self.ui.mca_plot_widget.update_spectrum)
        self.spectrum_changed.connect(self.ui.roi_plot_widget.update_spectrum)
        self.mca_row_hovered.connect(self.ui.mca_plot_widget.highlight_spectrum)
        self.roi_row_hovered.connect(self.ui.roi_plot_widget.highlight_spectrum)

    def enable_mca_checkboxes(self, new_state):
        """Check/uncheck the hinting checkboxes in response to the
        "Enable all" checkbox.

        """
        is_checked = new_state == Qt.Checked
        log.debug(f"Setting all MCA elements enabled: {new_state} / {is_checked}")
        self.ui.mca_enableall_checkbox.setTristate(False)
        self.enable_row_checkboxes(displays=self.mca_displays, is_checked=is_checked)

    def enable_roi_checkboxes(self, new_state):
        is_checked = new_state == Qt.Checked
        log.debug(f"Setting all ROI elements enabled: {new_state} / {is_checked}")
        self.ui.roi_enableall_checkbox.setTristate(False)
        self.enable_row_checkboxes(displays=self.roi_displays, is_checked=is_checked)

    def enable_row_checkboxes(self, displays, is_checked):
        """Check/uncheck the hinting checkboxes in response to the "Enable
        all" checkbox.

        """
        for display in displays:
            try:
                checkbox = display.embedded_widget.ui.enabled_checkbox
                checkbox.setChecked(is_checked)
            except AttributeError:
                pass

    def copy_selected_row(self, displays, source_display):
        """Copy the name, upper limit, and lower limits from one ROI|MCA row to all the rest.

        Parameters
        ==========
        display
          A sequence of all the displays to copy to.
        source_display
          The embedded display for the row from which to copy.

        """
        new_label = source_display.embedded_widget.ui.label_lineedit.text()
        new_lower = source_display.embedded_widget.ui.lower_lineedit.text()
        new_upper = source_display.embedded_widget.ui.upper_lineedit.text()
        # widget = source_display.embedded_widget.ui.upper_lineedit
        # Set all the other MCA rows with values from selected MCA row
        for display in displays:
            if display is not source_display:
                display.embedded_widget.ui.label_lineedit.setText(new_label)
                display.embedded_widget.ui.lower_lineedit.setText(new_lower)
                display.embedded_widget.ui.upper_lineedit.setText(new_upper)
                # Send the new value over the wire
                for widget in [
                    display.embedded_widget.ui.label_lineedit,
                    display.embedded_widget.ui.lower_lineedit,
                    display.embedded_widget.ui.upper_lineedit,
                ]:
                    if widget._connected:
                        widget.send_value()

    def copy_selected_roi(self):
        """Copy the label, hi channel, and lo channel values from the selected
        ROI row to all other visible ROI rows.

        """
        log.debug(f"Copying ROI {self._selected_roi}")
        # Get existing values from selected ROI row
        roi_idx = self._selected_roi
        source_display = self.roi_displays[roi_idx]
        self.copy_selected_row(
            source_display=source_display, displays=self.roi_displays
        )

    def copy_selected_mca(self):
        """Copy the label, hi channel, and lo channel values from the selected
        MCA row to all other visible MCA rows.

        """
        log.debug(f"Copying MCA {self._selected_mca}")
        # Get existing values from selected MCA row
        mca_idx = self._selected_mca - 1
        source_display = self.mca_displays[mca_idx]
        self.copy_selected_row(
            source_display=source_display, displays=self.mca_displays
        )

    def handle_new_spectrum(self, new_spectrum, mca_num):
        self.spectrum_changed.emit(mca_num, new_spectrum)

    def customize_device(self):
        # Load the device from the registry
        device_name = self.macros()["DEV"]
        self.device = device = haven.registry.find(device_name)
        # Set up data channels
        self._spectrum_channels = []
        for mca_num in range(1, self.device.num_elements + 1):
            address = f"oph://{device.name}.mcas.mca{mca_num}.spectrum"
            channel = pydm.PyDMChannel(
                address=address,
                value_slot=partial(self.handle_new_spectrum, mca_num=mca_num),
            )
            channel.connect()
            self._spectrum_channels.append(channel)

    def increment_combobox(self, combobox, step):
        n_items = combobox.count()
        new_index = (combobox.currentIndex() + step) % n_items
        combobox.setCurrentIndex(new_index)

    @contextmanager
    def disable_ui(self):
        widget = self
        # Disable to widgets
        widget.setEnabled(False)
        # Set waiting cursor
        old_cursor = self.cursor()
        self.setCursor(Qt.WaitCursor)
        # Update the UI
        FireflyApplication.instance().processEvents()
        yield
        # Re-enabled everything
        widget.setEnabled(True)
        widget.setCursor(old_cursor)

    def roi_selected(self, is_selected: bool, roi_num=None):
        """Handler for when an ROI is selected for editing.

        Parameters
        ==========
        is_selected
          Will be true if the ROI was selected, or false if the ROI
          was deselected.

        """
        # Save the selected ROI number for later
        if is_selected:
            self._selected_roi = roi_num
        else:
            self._selected_roi = None
        # Set global controls for this ROI
        self.ui.roi_copyall_button.setEnabled(is_selected)
        # Disable the other rows
        for idx, disp in enumerate(self.roi_displays):
            if is_selected and idx != roi_num:
                disp.setEnabled(False)
            else:
                disp.setEnabled(True)
        # Show this spectrum highlighted in the plots
        mca_num = self.ui.mca_combobox.currentIndex() + 1
        self.roi_plot_widget.select_roi(
            mca_num=mca_num, roi_num=roi_num, is_selected=is_selected
        )

    def mca_selected(self, is_selected: bool, mca_num: int = 0):
        """Handler for when an MCA row is selected for editing.

        Parameters
        ==========
        is_selected
          Will be true if the MCA row was selected, or false if the
          MCA row was deselected.

        """
        # Save the selected MCA number for later
        if is_selected:
            self._selected_mca = mca_num
        else:
            self._selected_mca = None
        # Set global controls for this MCA
        self.ui.mca_copyall_button.setEnabled(is_selected)
        # Disable the other rows
        mca_idx = mca_num - 1
        for idx, disp in enumerate(self.mca_displays):
            if is_selected and idx != mca_idx:
                disp.setEnabled(False)
            else:
                disp.setEnabled(True)
        # Show this spectrum highlighted in the plots
        roi_num = self.ui.roi_combobox.currentIndex()
        self.mca_plot_widget.select_spectrum(
            mca_num=mca_num, roi_num=roi_num, is_selected=is_selected
        )

    def draw_roi_widgets(self, mca_idx):
        mca_num = mca_idx + 1
        with self.disable_ui():
            # Update the plot widget with the new MCA number
            self.roi_plot_widget.target_mca = mca_num
            # Prepare all the ROI widgets
            layout = self.ui.rois_layout
            self.remove_widgets_from_layout(layout)
            self.roi_displays = []
            for roi_num in range(self.device.num_rois):
                disp = ROIEmbeddedDisplay(parent=self)
                disp.macros = json.dumps(
                    {
                        "DEV": self.device.name,
                        "MCA": mca_num,
                        "ROI": roi_num,
                        "NUM": roi_num,
                    }
                )
                disp.filename = "xrf_roi.py"
                # Respond when this ROI is selected
                disp.selected.connect(partial(self.roi_selected, roi_num=roi_num))
                disp.hovered.connect(
                    partial(self.roi_row_hovered.emit, mca_num, roi_num)
                )
                # Add the Embedded Display to the ROI Layout
                layout.addWidget(disp)
                self.roi_displays.append(disp)
            # Reset the selected ROI
            self.roi_selected(is_selected=False)
            # Make the global enable checkbox tri-state again
            self.ui.roi_enableall_checkbox.setCheckState(Qt.PartiallyChecked)

    def draw_mca_widgets(self, roi_idx):
        """Prepare a row for each element in the detector."""
        with self.disable_ui():
            # Prepare all the ROI widgets
            layout = self.ui.mcas_layout
            self.remove_widgets_from_layout(layout)
            self.mca_displays = []
            for mca_num in range(1, self.device.num_elements + 1):
                disp = ROIEmbeddedDisplay(parent=self)
                disp.macros = json.dumps(
                    {
                        "DEV": self.device.name,
                        "MCA": mca_num,
                        "ROI": roi_idx,
                        "NUM": mca_num,
                    }
                )
                disp.filename = "xrf_roi.py"
                # Respond when this MCA is interacted with
                disp.selected.connect(partial(self.mca_selected, mca_num=mca_num))
                disp.hovered.connect(
                    partial(self.mca_row_hovered.emit, mca_num, roi_idx)
                )
                # Add the Embedded Display to the ROI Layout
                layout.addWidget(disp)
                self.mca_displays.append(disp)
            # Reset the selected MCA
            self.mca_selected(is_selected=False)
            # Make the global enable checkbox tri-state again
            self.ui.mca_enableall_checkbox.setCheckState(Qt.PartiallyChecked)

    def remove_widgets_from_layout(self, layout):
        # Delete existing ROI widgets
        for idx in reversed(range(layout.count())):
            layout.takeAt(idx).widget().deleteLater()



import asyncio
import string
import sys
from datetime import timedelta

import numpy
import numpy as np
import pandas
import pandas as pd
import sparse
import xarray

from tiled.adapters.array import ArrayAdapter
from tiled.adapters.dataframe import DataFrameAdapter
from tiled.adapters.mapping import MapAdapter
from tiled.adapters.sparse import COOAdapter
from tiled.adapters.xarray import DatasetAdapter

# print("Generating large example data...", file=sys.stderr)
# data = {
#     "big_image": numpy.random.random((10_000, 10_000)),
#     "small_image": numpy.random.random((300, 300)),
#     "medium_image": numpy.random.random((1000, 1000)),
#     "tiny_image": numpy.random.random((50, 50)),
#     "tiny_cube": numpy.random.random((50, 50, 50)),
#     "tiny_hypercube": numpy.random.random((50, 50, 50, 50, 50)),
#     "high_entropy": numpy.random.random((100, 100)),
#     "low_entropy": numpy.ones((100, 100)),
#     "short_column": numpy.random.random(100),
#     "tiny_column": numpy.random.random(10),
#     "long_column": numpy.random.random(100_000),
# }
# temp = 15 + 8 * numpy.random.randn(2, 2, 3)
# precip = 10 * numpy.random.rand(2, 2, 3)
# lon = [[-99.83, -99.32], [-99.79, -99.23]]
# lat = [[42.25, 42.21], [42.63, 42.59]]

# sparse_arr = numpy.random.random((100, 100))
# sparse_arr[sparse_arr < 0.9] = 0  # fill most of the array with zeros

run1 = pd.DataFrame(
    {
        "energy_energy": np.linspace(8300, 8400, num=100),
        "It_net_counts": np.abs(np.sin(np.linspace(0, 4 * np.pi, num=100))),
        "I0_net_counts": np.linspace(1, 2, num=100),
    }
)

print("Done generating example data.", file=sys.stderr)
hints = {
    "energy": {"fields": ["energy_energy", "energy_id_energy_readback"]},
}

bluesky_mapping = {
    "7d1daf1d-60c7-4aa7-a668-d1cd97e5335f": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(run1.to_xarray()),
                },
                metadata={"descriptors": [{"hints": hints}]},
            ),
        },
        metadata={
            "plan_name": "xafs_scan",
            "start": {
                "plan_name": "xafs_scan",
                "uid": "7d1daf1d-60c7-4aa7-a668-d1cd97e5335f",
                "hints": {"dimensions": [[["energy_energy"], "primary"]]},
            },
        },
    ),
    "9d33bf66-9701-4ee3-90f4-3be730bc226c": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(run1.to_xarray()),
                },
                metadata={"descriptors": [{"hints": hints}]},
            ),
        },
        metadata={
            "start": {
                "plan_name": "rel_scan",
                "uid": "9d33bf66-9701-4ee3-90f4-3be730bc226c",
                "hints": {"dimensions": [[["pitch2"], "primary"]]},
            }
        },
    ),
    # '1942a888-2627-43e6-ad36-82f0022e2c57': MapAdapter({
    #     "start": {
    #         "plan_name": "xafs_scan"
    #     }
    # }),
    # 'bb1cd731-f180-40b5-8255-71a6a398e51c': MapAdapter({
    #     "start": {
    #         "plan_name": "xafs_scan"
    #     }
    # }),
    # '4165b2bb-7df5-4a0e-8c3e-8221a9808bef': MapAdapter({
    #     "start": {
    #         "plan_name": "xafs_scan"
    #     }
    # }),
    # 'e3131e5e-4ebf-458e-943b-62bcdbc0f6e0': MapAdapter({
    #     "start": {
    #         "plan_name": "xafs_scan"
    #     }
    # }),
}

# mapping = {
#     "big_image": ArrayAdapter.from_array(data["big_image"]),
#     "small_image": ArrayAdapter.from_array(data["small_image"]),
#     "medium_image": ArrayAdapter.from_array(data["medium_image"]),
#     "sparse_image": COOAdapter.from_coo(sparse.COO(sparse_arr)),
#     "tiny_image": ArrayAdapter.from_array(data["tiny_image"]),
#     "tiny_cube": ArrayAdapter.from_array(data["tiny_cube"]),
#     "tiny_hypercube": ArrayAdapter.from_array(data["tiny_hypercube"]),
#     "short_table": DataFrameAdapter.from_pandas(
#         pandas.DataFrame(
#             {
#                 "A": data["short_column"],
#                 "B": 2 * data["short_column"],
#                 "C": 3 * data["short_column"],
#             },
#             index=pandas.Index(numpy.arange(len(data["short_column"])), name="index"),
#         ),
#         npartitions=1,
#         metadata={"animal": "dog", "color": "red"},
#     ),
#     "long_table": DataFrameAdapter.from_pandas(
#         pandas.DataFrame(
#             {
#                 "A": data["long_column"],
#                 "B": 2 * data["long_column"],
#                 "C": 3 * data["long_column"],
#             },
#             index=pandas.Index(numpy.arange(len(data["long_column"])), name="index"),
#         ),
#         npartitions=5,
#         metadata={"animal": "dog", "color": "green"},
#     ),
#     "wide_table": DataFrameAdapter.from_pandas(
#         pandas.DataFrame(
#             {
#                 letter: i * data["tiny_column"]
#                 for i, letter in enumerate(string.ascii_uppercase, start=1)
#             },
#             index=pandas.Index(numpy.arange(len(data["tiny_column"])), name="index"),
#         ),
#         npartitions=1,
#         metadata={"animal": "dog", "color": "red"},
#     ),
#     "structured_data": MapAdapter(
#         {
#             "pets": ArrayAdapter.from_array(
#                 numpy.array(
#                     [("Rex", 9, 81.0), ("Fido", 3, 27.0)],
#                     dtype=[("name", "U10"), ("age", "i4"), ("weight", "f4")],
#                 )
#             ),
#             "xarray_dataset": DatasetAdapter.from_dataset(
#                 xarray.Dataset(
#                     {
#                         "temperature": (["x", "y", "time"], temp),
#                         "precipitation": (["x", "y", "time"], precip),
#                     },
#                     coords={
#                         "lon": (["x", "y"], lon),
#                         "lat": (["x", "y"], lat),
#                         "time": pandas.date_range("2014-09-06", periods=3),
#                     },
#                 ),
#             ),
#         },
#         metadata={"animal": "cat", "color": "green"},
#     ),
#     "flat_array": ArrayAdapter.from_array(numpy.random.random(100)),
#     "low_entropy": ArrayAdapter.from_array(data["low_entropy"]),
#     "high_entropy": ArrayAdapter.from_array(data["high_entropy"]),
#     # Below, an asynchronous task modifies this value over time.
#     "dynamic": ArrayAdapter.from_array(numpy.zeros((3, 3))),
# }

mapping = {
    "255id_testing": MapAdapter(bluesky_mapping),
}

tree = MapAdapter(mapping, entries_stale_after=timedelta(seconds=10))


# async def increment_dynamic():
#     """
#     Change the value of the 'dynamic' node every 3 seconds.
#     """
#     fill_value = 0
#     while True:
#         fill_value += 1
#         mapping["dynamic"] = ArrayAdapter.from_array(fill_value * numpy.ones((3, 3)))
#         await asyncio.sleep(3)


# # The server will run this on its event loop. We cannot start it *now* because
# # there is not yet a running event loop.
# tree.background_tasks.append(increment_dynamic)



import time
import pytest

import matplotlib.pyplot as plt
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.callbacks.fitting import PeakStats
from ophyd import sim

from haven import align_motor, align_pitch2, registry
from run_engine import RunEngineStub


def test_align_motor(ffapp):
    # Set up simulated motors and detectors
    motor = sim.SynAxis(name="motor", labels={"motors"})
    detector = sim.SynGauss(
        name="detector",
        motor=motor,
        motor_field="motor",
        center=-3,
        Imax=1,
        sigma=20,
        labels={"detectors"},
    )
    # Prepare the callback to check results
    bec = BestEffortCallback()
    bec.disable_plots()
    bec.disable_table()
    # Prepare the plan
    plan = align_motor(
        detector=detector,
        motor=motor,
        distance=40,
        bec=bec,
        md={"plan_name": "test_plan"},
    )
    # Execute the plan
    RE = RunEngineStub(call_returns_result=True)
    result = RE(plan)
    # Check peak calculation results
    assert bec.peaks["cen"]["detector"] == pytest.approx(-3, rel=1e-3)
    assert motor.readback.get() == pytest.approx(-3, rel=1e-3)


# def test_align_pitch2():
#     # Prepare fake motors
#     motor = registry.register(sim.SynAxis(name="monochromator_pitch2", labels={"motors"}))
#     detector = registry.register(sim.SynGauss(
#         name="I0",
#         motor=motor,
#         motor_field="monochromator_pitch2",
#         center=-3,
#         Imax=1,
#         sigma=20,
#         labels={"detectors"},
#     ))
#     # Prepare the callback to check results
#     bec = BestEffortCallback()
#     bec.disable_plots()
#     bec.disable_table()
#     # Prepare the plan
#     plan = align_pitch2(bec=bec, distance=40, reverse=True)
#     # Execute the plan
#     RE = RunEngineStub(call_returns_result=True)
#     result = RE(plan)
#     # Check that the pitch motor gets moved
#     assert "I0" in bec.peaks['cen'].keys()
#     assert bec.peaks['cen']['I0'] == pytest.approx(-3, rel=1e-3)
#     assert motor.readback.get() == pytest.approx(-3, rel=1e-3)



import gc

import databroker
from bluesky import RunEngine

from haven.instrument.aps import load_aps
from haven import run_engine


def test_subscribers_garbage_collection(monkeypatch):
    """Tests for regression of a bug in databroker.

    Since databroker uses a weak reference to the insert function, it
    can be subject to garbage collection and no longer able to save
    data.

    """
    monkeypatch.setattr(databroker, "catalog", {"bluesky": databroker.temp()})
    load_aps()
    RE = run_engine()
    assert len(RE.dispatcher.cb_registry.callbacks) == 12
    gc.collect()
    assert len(RE.dispatcher.cb_registry.callbacks) == 12


def test_run_engine_preprocessors():
    load_aps()
    RE = run_engine()
    assert len(RE.preprocessors) > 0


def test_run_engine_created():
    load_aps()
    RE = run_engine()
    assert isinstance(RE, RunEngine)



from ophyd import PVPositioner, EpicsSignalRO, EpicsSignalWithRBV, Component as Cpt
from apstools.devices import (
    PTC10PositionerMixin,
    PTC10AioChannel as PTC10AioChannelBase,
    PTC10TcChannel,
)

from .._iconfig import load_config
from .instrument_registry import registry


# The apstools version uses "voltage_RBV" as the PVname
class PTC10AioChannel(PTC10AioChannelBase):
    """
    SRS PTC10 AIO module
    """

    voltage = Cpt(EpicsSignalRO, "output_RBV", kind="config")


@registry.register
class CapillaryHeater(PTC10PositionerMixin, PVPositioner):
    readback = Cpt(EpicsSignalRO, "2A:temperature", kind="hinted")
    setpoint = Cpt(EpicsSignalWithRBV, "5A:setPoint", kind="hinted")

    # Additional modules installed on the PTC10
    pid = Cpt(PTC10AioChannel, "5A:")
    tc = Cpt(PTC10TcChannel, "2A:")


def load_heaters(config=None):
    if config is None:
        config = load_config()
    # Load the heaters
    heaters = []
    for name, cfg in config["heater"].items():
        Cls = globals().get(cfg["device_class"])
        device = Cls(prefix=f"{cfg['prefix']}:", name=name)
        heaters.append(device)
    return heaters



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import PVGroup, SubGroup, ioc_arg_parser, pvproperty, run
from ophyd.tests.fake_motor_ioc import FakeMotorIOC

from haven.simulated_ioc import ResponsiveMotorFields  # , IOC as IOC_


class MonoGroup(PVGroup):
    """
    An IOC with a monochromator.

    """

    m1 = pvproperty(value=0, doc="horiz", name="ACS:m1", record=ResponsiveMotorFields)
    m2 = pvproperty(value=0, doc="vert", name="ACS:m2", record=ResponsiveMotorFields)
    m3 = pvproperty(value=0, doc="bragg", name="ACS:m3", record=ResponsiveMotorFields)
    m4 = pvproperty(value=0, doc="gap", name="ACS:m4", record=ResponsiveMotorFields)
    m5 = pvproperty(value=0, doc="roll2", name="ACS:m5", record=ResponsiveMotorFields)
    m6 = pvproperty(value=0, doc="pitch2", name="ACS:m6", record=ResponsiveMotorFields)
    m7 = pvproperty(
        value=0, doc="roll-int", name="ACS:m7", record=ResponsiveMotorFields
    )
    m8 = pvproperty(value=0, doc="pi-int", name="ACS:m8", record=ResponsiveMotorFields)
    Energy = pvproperty(value=10000.0, doc="Energy", record=ResponsiveMotorFields)
    Offset = pvproperty(value=9009, doc="Offset", record=ResponsiveMotorFields)
    mode = pvproperty(value=1, doc="mode", record=ResponsiveMotorFields)
    id_tracking = pvproperty(value=0, name="ID_tracking", doc="IDTracking")
    id_offset = pvproperty(value=0, name="ID_offset", doc="ID offset")
    dspacing = pvproperty(value=3.135, doc="Crystal d-spacing", dtype=float)
    EnergyC1 = pvproperty(value=1, name="EnergyC1.VAL", dtype=float)
    EnergyC2 = pvproperty(value=1, name="EnergyC2.VAL", dtype=float)
    EnergyC3 = pvproperty(value=1, name="EnergyC3.VAL", dtype=float)


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idMono:", desc="haven.tests.ioc_mono test IOC"
    )
    ioc = MonoGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import (
    PVGroup,
    SubGroup,
    ioc_arg_parser,
    pvproperty,
    run,
    PvpropertyString,
    PvpropertyInteger,
)
from ophyd.tests.fake_motor_ioc import FakeMotorIOC

from haven.simulated_ioc import ResponsiveMotorFields  # , IOC as IOC_


class PreampsGroup(PVGroup):
    class PreampGroup(PVGroup):
        """An IOC with an SR570 pre-amplifier.

        E.g.

        - 25idc:SR01:IpreSlit:sens_num.VAL"
        - 25idc:SR01:IpreSlit:sens_unit.VAL

        """

        sens_num = pvproperty(
            value="5",
            enum_strings=["1", "2", "5", "10", "20", "50", "100", "200", "500"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        sens_unit = pvproperty(
            value="nA/V",
            enum_strings=["pA/V", "nA/V", "uA/V", "mA/V"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        offset_num = pvproperty(
            value="5",
            enum_strings=["1", "2", "5", "10", "20", "50", "100", "200", "500"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        offset_unit = pvproperty(
            value="nA",
            enum_strings=["pA", "nA", "uA", "mA"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        set_all = pvproperty(value=1, name="init.PROC", doc="")

    preamp1 = SubGroup(PreampGroup, prefix="SR01:")
    preamp2 = SubGroup(PreampGroup, prefix="SR02:")
    preamp3 = SubGroup(PreampGroup, prefix="SR03:")
    preamp4 = SubGroup(PreampGroup, prefix="SR04:")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idc:", desc="haven.tests.ioc_preamp test IOC"
    )
    ioc = PreampsGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



from haven.catalog import tiled_client


def test_tiled_client(sim_tiled):
    uri = sim_tiled.uri.split("/")[2]
    uri = f"http://{uri}"
    client = tiled_client(entry_node="255id_testing", uri=uri)
    assert "7d1daf1d-60c7-4aa7-a668-d1cd97e5335f" in client.keys()
    run = client["7d1daf1d-60c7-4aa7-a668-d1cd97e5335f"]
    assert run.metadata["start"]["plan_name"] == "xafs_scan"
